"""
Module provides a class for implementing sensor configurations.
"""
from copy import deepcopy
import itertools
import numpy as np
from epanet_plus import EpanetConstants, EPyT

from ..serialization import SENSOR_CONFIG_ID, JsonSerializable, serializable


SENSOR_TYPE_NODE_PRESSURE          = 1
SENSOR_TYPE_NODE_QUALITY           = 2
SENSOR_TYPE_NODE_DEMAND            = 3
SENSOR_TYPE_LINK_FLOW              = 4
SENSOR_TYPE_LINK_QUALITY           = 5
SENSOR_TYPE_VALVE_STATE            = 6
SENSOR_TYPE_PUMP_STATE             = 7
SENSOR_TYPE_TANK_VOLUME            = 8
SENSOR_TYPE_NODE_BULK_SPECIES      = 9
SENSOR_TYPE_LINK_BULK_SPECIES      = 10
SENSOR_TYPE_SURFACE_SPECIES        = 11
SENSOR_TYPE_PUMP_EFFICIENCY        = 12
SENSOR_TYPE_PUMP_ENERGYCONSUMPTION = 13

AREA_UNIT_FT2 = 1
AREA_UNIT_M2 = 2
AREA_UNIT_CM2 = 3
MASS_UNIT_MG = 4
MASS_UNIT_UG = 5
MASS_UNIT_MOL = 6
MASS_UNIT_MMOL = 7
TIME_UNIT_HRS = 8
MASS_UNIT_CUSTOM = 9


def areaunit_to_id(unit_desc: str) -> int:
    """
    Converts a given area units string to the corresponding ID.

    Parameters
    ----------
    unit_desc : `str`
        Area units string.

    Returns
    -------
    `int`
        Corresponding area unit ID.
    """
    return {"FT2": AREA_UNIT_FT2,
            "M2": AREA_UNIT_M2,
            "CM2": AREA_UNIT_CM2}[unit_desc]


def massunit_to_id(unit_desc: str) -> int:
    """
    Converts a given mass units string to the corresponding ID.

    Parameters
    ----------
    unit_desc : `str`
        Mass units string.

    Returns
    -------
    `int`
        Corresponding mass unit ID.
    """
    mass_unit_dict = {"MG": MASS_UNIT_MG,
                      "UG": MASS_UNIT_UG,
                      "MOL": MASS_UNIT_MOL,
                      "MMOL": MASS_UNIT_MMOL}

    if unit_desc in mass_unit_dict:
        return mass_unit_dict[unit_desc]
    else:
        return MASS_UNIT_CUSTOM


def qualityunit_to_id(unit_desc: str) -> int:
    """
    Converts a given measurement unit description to the corresponding mass unit ID.

    Parameters
    ----------
    unit_desc : `str`
        Mass unit.

    Returns
    -------
    `int`
        Mass unit ID.

        Will be either None (if no water quality analysis was set up) or
        one of the following constants:

            - MASS_UNIT_MG  = 4   (mg/L)
            - MASS_UNIT_UG  = 5   (ug/L)
            - TIME_UNIT_HRS = 8  (hrs)
    """
    if unit_desc == "mg/L":
        return MASS_UNIT_MG
    elif unit_desc == "ug/L":
        return MASS_UNIT_UG
    elif unit_desc == "hrs":
        return TIME_UNIT_HRS
    else:
        return None


def massunit_to_str(unit_id: int) -> str:
    """
    Converts a given mass unit ID to the corresponding description.

    Parameters
    ----------
    unit_id : `int`
        ID of the mass unit.

        Must be one of the following constant:

            - MASS_UNIT_MG     = 4
            - MASS_UNIT_UG     = 5
            - MASS_UNIT_MOL    = 6
            - MASS_UNIT_MMOL   = 7
            - MASS_UNIT_CUSTOM = 9

    Returns
    -------
    `str`
        Mass unit description.
    """
    if unit_id is None:
        return ""
    elif unit_id == MASS_UNIT_MG:
        return "MG"
    elif unit_id == MASS_UNIT_UG:
        return "UG"
    elif unit_id == MASS_UNIT_MOL:
        return "MOL"
    elif unit_id == MASS_UNIT_MMOL:
        return "MMOL"
    elif unit_id == MASS_UNIT_CUSTOM:
        return "CUSTOM UNIT"
    else:
        raise ValueError(f"Unknown mass unit ID '{unit_id}'")


def flowunit_to_str(unit_id: int) -> str:
    """
    Converts a given flow unit ID to the corresponding description.

    Parameters
    ----------
    unit_id : `int`
        ID of the flow unit.

        Must be one of the following EPANET constants:

            - EN_CFS  = 0 (cubic foot/sec)
            - EN_GPM  = 1 (gal/min)
            - EN_MGD  = 2 (Million gal/day)
            - EN_IMGD = 3 (Imperial MGD)
            - EN_AFD  = 4 (ac-foot/day)
            - EN_LPS  = 5 (liter/sec)
            - EN_LPM  = 6 (liter/min)
            - EN_MLD  = 7 (Megaliter/day)
            - EN_CMH  = 8 (cubic meter/hr)
            - EN_CMD  = 9 (cubic meter/day)

    Returns
    -------
    `str`
        Flow unit description.
    """
    if unit_id is None:
        return ""
    elif unit_id == EpanetConstants.EN_CFS:
        return "cubic foot/sec"
    elif unit_id == EpanetConstants.EN_GPM:
        return "gal/min"
    elif unit_id == EpanetConstants.EN_MGD:
        return "Million gal/day"
    elif unit_id == EpanetConstants.EN_IMGD:
        return "Imperial MGD"
    elif unit_id == EpanetConstants.EN_AFD:
        return "ac-foot/day"
    elif unit_id == EpanetConstants.EN_LPS:
        return "liter/sec"
    elif unit_id == EpanetConstants.EN_LPM:
        return "liter/min"
    elif unit_id == EpanetConstants.EN_MLD:
        return "Megaliter/day"
    elif unit_id == EpanetConstants.EN_CMH:
        return "cubic meter/hr"
    elif unit_id == EpanetConstants.EN_CMD:
        return "cubic meter/day"
    else:
        raise ValueError(f"Unknown unit ID '{unit_id}'")


def qualityunit_to_str(unit_id: int) -> str:
    """
    Converts a given quality measurement unit ID to the corresponding description.

    Parameters
    ----------
    unit_id : `int`
        ID of the quality unit.

        Must be one of the following constants:

            - MASS_UNIT_MG  = 4  (mg/L)
            - MASS_UNIT_UG  = 5  (ug/L)
            - TIME_UNIT_HRS = 8  (hrs)

    Returns
    -------
    `str`
        Mass unit description.
    """
    if unit_id is None:
        return ""
    elif unit_id == MASS_UNIT_MG:
        return "mg/L"
    elif unit_id == MASS_UNIT_UG:
        return "ug/L"
    elif unit_id == TIME_UNIT_HRS:
        return "hrs"
    else:
        raise ValueError(f"Unknown unit ID '{unit_id}'")


def areaunit_to_str(unit_id: int) -> str:
    """
    Converts a given area measurement unit ID to the corresponding description.

    Parameters
    ----------
    unit_id : `int`
        ID of the area unit.

        Must be one of the following constants:

            - AREA_UNIT_FT2 = 1
            - AREA_UNIT_M2  = 2
            - AREA_UNIT_CM2 = 3

    Returns
    -------
    `str`
        Area unit description.
    """
    if unit_id is None:
        return None
    elif unit_id == AREA_UNIT_FT2:
        return "FT2"
    elif unit_id == AREA_UNIT_M2:
        return "M2"
    elif unit_id == AREA_UNIT_CM2:
        return "CM2"
    else:
        raise ValueError(f"Unknown unit ID '{unit_id}'")


def is_flowunit_simetric(unit_id: int) -> bool:
    """
    Checks if a given flow unit belongs to SI metric units.

    Parameters
    ----------
    unit_id : `int`
        ID of the flow unit.

        Must be one of the following EPANET constants:

            - EN_CFS  = 0 (cubic foot/sec)
            - EN_GPM  = 1 (gal/min)
            - EN_MGD  = 2 (Million gal/day)
            - EN_IMGD = 3 (Imperial MGD)
            - EN_AFD  = 4 (ac-foot/day)
            - EN_LPS  = 5 (liter/sec)
            - EN_LPM  = 6 (liter/min)
            - EN_MLD  = 7 (Megaliter/day)
            - EN_CMH  = 8 (cubic meter/hr)
            - EN_CMD  = 9 (cubic meter/day)

    Returns
    -------
    `bool`
        True if the fiven unit is a SI metric unit, False otherwise.
    """
    return unit_id in [EpanetConstants.EN_LPS, EpanetConstants.EN_LPM, EpanetConstants.EN_MLD,
                       EpanetConstants.EN_CMH, EpanetConstants.EN_CMD]


@serializable(SENSOR_CONFIG_ID, ".epytflow_sensor_config")
class SensorConfig(JsonSerializable):
    """
    Class for storing a sensor configuration.

    Parameters
    ----------
    nodes : `list[str]`
        List of all nodes (i.e. IDs) in the network.
    links : `list[str]`
        List of all links/pipes (i.e. IDs) in the network.
    valves : `list[str]`
        List of all valves (i.e. IDs) in the network.
    pumps : `list[str]`
        List of all pumps (i.e. IDs) in the network.
    tanks : `list[str]`
        List of all tanks (i.e. IDs) in the network.
    species : `list[str]`
        List of all (EPANET-MSX) species (i.e. IDs) in the network
    pressure_sensors : `list[str]`, optional
        List of all nodes (i.e. IDs) at which a pressure sensor is placed.

        The default is an empty list.
    flow_sensors : `list[str]`, optional
        List of all links/pipes (i.e. IDs) at which a flow sensor is placed.

        The default is an empty list.
    demand_sensors : `list[str]`, optional
        List of all nodes (i.e. IDs) at which a demand sensor is placed.

        The default is an empty list.
    quality_node_sensors : `list[str]`, optional
        List of all nodes (i.e. IDs) at which a quality sensor is placed.

        The default is an empty list.
    quality_link_sensors : `list[str]`, optional
        List of all links/pipes (i.e. IDs) at which a flow sensor is placed.

        The default is an empty list.
    valve_state_sensors : `list[str]`, optional
        List of all valves (i.e. IDs) at which a valve state sensor is placed.

        The default is an empty list.
    pump_state_sensors : `list[str]`, optional
        List of all pumps (i.e. IDs) at which a pump state sensor is placed.

        The default is an empty list.
    tank_volume_sensors : `list[str]`, optional
        List of all tanks (i.e. IDs) at which a tank volume sensor is placed.

        The default is an empty list.
    bulk_species_node_sensors : `dict`, optional
        Bulk species node sensors as a dictionary -- i.e. bulk species ID are the keys,
        and the sensor locations (node IDs) are the values.

        The default is an empty list.
    bulk_species_link_sensors : `dict`, optional
        Bulk species link/pipe sensors as a dictionary -- i.e. bulk species ID are the keys,
        and the sensor locations (link/pipe IDs) are the values.

        The default is an empty list.
    surface_species_sensors : `dict`, optional
        Surface species sensors as a dictionary -- i.e. surface species ID are the keys,
        and the sensor locations (link/pipe IDs) are the values.

        The default is an empty list.
    node_id_to_idx : `dict`, optional
        Mapping of a node ID to the EPANET index (i.e. position in the raw sensor reading data).

        If None is given, it is assumed that the nodes (in 'nodes') are
        sorted according to their EPANET index.

        The default is None.
    link_id_to_idx : `dict`, optional
        Mapping of a link/pipe ID to the EPANET index
        (i.e. position in the raw sensor reading data).

        If None is given, it is assumed that the links/pipes (in 'links') are
        sorted according to their EPANET index..

        The default is None.
    valve_id_to_idx : `dict`, optional
        Mapping of a valve ID to the EPANET index (i.e. position in the raw sensor reading data).

        If None is given, it is assumed that the valves (in 'valves') are
        sorted according to their EPANET index.

        The default is None.
    pump_id_to_idx : `dict`, optional
        Mapping of a pump ID to the EPANET index (i.e. position in the raw sensor reading data).

        If None is given, it is assumed that the pumps (in 'pumps') are
        sorted according to their EPANET index.

        The default is None.
    tank_id_to_idx : `dict`, optional
        Mapping of a tank ID to the EPANET index (i.e. position in the raw sensor reading data).

        If None is given, it is assumed that the tanks (in 'tanks') are
        sorted according to their EPANET index.

        The default is None.
    bulkspecies_id_to_idx : `dict`, optional
        Mapping of a surface species ID to the EPANET index
        (i.e. position in the raw sensor reading data).

        If None is given, it is assumed that the surface species (in 'surface_species') are
        sorted according to their EPANET index.

        The default is None.
    flow_unit : `int`
        Specifies the flow units and consequently all other hydraulic units
        (US CUSTOMARY or SI METRIC) as well.

        Must be one of the following EPANET constants:

            - EN_CFS = 0 (cubic foot/sec)
            - EN_GPM = 1 (gal/min)
            - EN_MGD = 2 (Million gal/day)
            - EN_IMGD = 3 (Imperial MGD)
            - EN_AFD = 4 (ac-foot/day)
            - EN_LPS = 5 (liter/sec)
            - EN_LPM = 6 (liter/min)
            - EN_MLD = 7 (Megaliter/day)
            - EN_CMH = 8 (cubic meter/hr)
            - EN_CMD = 9 (cubic meter/day)
    quality_unit : `str`, optional
        Measurement unit (in a basic quality analysis) -- only relevant
        if basic water quality is enabled.

        Must be one of the following constants:

            - MASS_UNIT_MG = 4     (mg/L)
            - MASS_UNIT_UG = 5     (ug/L)
            - TIME_UNIT_HRS = 8    (hrs)

    bulk_species_mass_unit : `list[int]`, optional
        Specifies the mass unit for each bulk species -- only relevant if EPANET-MSX is used.

        Must be one of the following constants:

            - MASS_UNIT_MG = 4      (milligram)
            - MASS_UNIT_UG = 5      (microgram)
            - MASS_UNIT_MOL = 6     (mole)
            - MASS_UNIT_MMOL = 7    (millimole)

        Note that the assumed ordering is the same as given in 'bulk_species'.
    surface_species_mass_unit : `list[int]`, optional
        Specifies the mass unit for each surface species -- only relevant if EPANET-MSX is used.

        Must be one of the following constants:

            - MASS_UNIT_MG = 4      (milligram)
            - MASS_UNIT_UG = 5      (microgram)
            - MASS_UNIT_MOL = 6     (mole)
            - MASS_UNIT_MMOL = 7    (millimole)

        Note that the assumed ordering is the same as given in 'surface_species'.
    surface_species_area_unit : `int`, optional
        Species the area unit of all surface species -- only relevant if EPANET-MSX is used.
        Must be one of the following constants:

            - AREA_UNIT_FT2 = 1     (square feet)
            - AREA_UNIT_M2 = 2      (square meters)
            - AREA_UNIT_CM2 = 3     (square centimeters)
    """
    def __init__(self, nodes: list[str], links: list[str], valves: list[str], pumps: list[str],
                 tanks: list[str], bulk_species: list[str], surface_species: list[str],
                 flow_unit: int,
                 pressure_sensors: list[str] = [],
                 flow_sensors: list[str] = [],
                 demand_sensors: list[str] = [],
                 quality_node_sensors: list[str] = [],
                 quality_link_sensors: list[str] = [],
                 valve_state_sensors: list[str] = [],
                 pump_state_sensors: list[str] = [],
                 pump_efficiency_sensors: list[str] = [],
                 pump_energyconsumption_sensors: list[str] = [],
                 tank_volume_sensors: list[str] = [],
                 bulk_species_node_sensors: dict = {},
                 bulk_species_link_sensors: dict = {},
                 surface_species_sensors: dict = {},
                 node_id_to_idx: dict = None, link_id_to_idx: dict = None,
                 valve_id_to_idx: dict = None, pump_id_to_idx: dict = None,
                 tank_id_to_idx: dict = None, bulkspecies_id_to_idx: dict = None,
                 surfacespecies_id_to_idx: dict = None,
                 quality_unit: int = None,
                 bulk_species_mass_unit : list[int] = [],
                 surface_species_mass_unit : list[int] = [],
                 surface_species_area_unit : int = None,
                 **kwds):
        if not isinstance(nodes, list):
            raise TypeError("'nodes' must be an instance of 'list[str]' " +
                            f"but not of '{type(nodes)}'")
        if len(nodes) == 0:
            raise ValueError("'nodes' must be a list of all nodes (i.e. IDs) in the network.")
        if any(not isinstance(n, str) for n in nodes):
            raise TypeError("Each item in 'nodes' must be an instance of 'str' -- " +
                            "ID of a node in the network.")

        if not isinstance(links, list):
            raise TypeError("'links' must be an instance of 'list[str]' " +
                            f"but not of '{type(links)}'")
        if len(links) == 0:
            raise ValueError("'links' must be a list of all links/pipes (i.e. IDs) in the network.")
        if any(not isinstance(link, str) for link in links):
            raise TypeError("Each item in 'links' must be an instance of 'str' -- " +
                            "ID of a link/pipe in the network.")

        if not isinstance(valves, list):
            raise TypeError("'valves' must be an instance of 'list[str]' " +
                            f"but not of '{type(valves)}'")
        if any(v not in links for v in valves):
            raise ValueError("Each item in 'valves' must be in 'links'")

        if not isinstance(pumps, list):
            raise TypeError("'pumps' must be an instance of 'list[str]' " +
                            f"but not of '{type(pumps)}'")
        if any(p not in links for p in pumps):
            raise ValueError("Each item in 'pumps' must be in 'links'")

        if not isinstance(tanks, list):
            raise TypeError("'tanks' must be an instance of 'list[str]' " +
                            f"but not of '{type(tanks)}'")
        if any(v not in nodes for v in tanks):
            raise ValueError("Each item in 'tanks' must be in 'nodes'")

        if not isinstance(bulk_species, list):
            raise TypeError("'bulk_species' must be an instance of 'list[str]' " +
                            f"but not of '{type(bulk_species)}'")
        if any(not isinstance(bulk_species_id, str) for bulk_species_id in bulk_species):
            raise TypeError("Each item in 'bulk_species' must be an instance of 'str'")

        if not isinstance(surface_species, list):
            raise TypeError("'surface_species' must be an instance of 'list[str]' " +
                            f"but not of '{type(surface_species)}'")
        if any(not isinstance(surface_species_id, str) for surface_species_id in surface_species):
            raise TypeError("Each item in 'surface_species' must be an instance of 'str'")

        if not isinstance(pressure_sensors, list):
            raise TypeError("'pressure_sensors' must be an instance of 'list[str]' " +
                            f"but not of '{type(pressure_sensors)}'")
        if any(n not in nodes for n in pressure_sensors):
            raise ValueError("Each item in 'pressure_sensors' must be in 'nodes' -- " +
                             "cannot place a sensor at a non-existing node.")

        if not isinstance(flow_sensors, list):
            raise TypeError("'flow_sensors' must be an instance of 'list[str]' " +
                            f"but not of '{type(flow_sensors)}'")
        if any(link not in links for link in flow_sensors):
            raise ValueError("Each item in 'flow_sensors' must be in 'links' -- cannot " +
                             "place a sensor at a non-existing link/pipe.")

        if not isinstance(demand_sensors, list):
            raise TypeError("'demand_sensors' must be an instance of 'list[str]' " +
                            f"but not of '{type(demand_sensors)}'")
        if any(n not in nodes for n in demand_sensors):
            raise ValueError("Each item in 'demand_sensors' must be in 'nodes' -- cannot " +
                             "place a sensor at a non-existing node.")

        if not isinstance(quality_node_sensors, list):
            raise TypeError("'quality_node_sensors' must be an instance of 'list[str]' " +
                            f"but not of '{type(quality_node_sensors)}'")
        if any(n not in nodes for n in quality_node_sensors):
            raise ValueError("Each item in 'quality_node_sensors' must be in 'nodes' -- cannot " +
                             "place a sensor at a non-existing node.")

        if not isinstance(quality_link_sensors, list):
            raise TypeError("'quality_link_sensors' must be an instance of 'list[str]' " +
                            f"but not of '{type(quality_link_sensors)}'")
        if any(link not in links for link in quality_link_sensors):
            raise ValueError("Each item in 'quality_link_sensors' must be in 'links' -- cannot " +
                             "place a sensor at a non-existing link/pipe.")

        if not isinstance(valve_state_sensors, list):
            raise TypeError("'valve_state_sensors' must be an instance of 'list[str]' " +
                            f"but not of '{type(valve_state_sensors)}'")
        if any(link not in valves for link in valve_state_sensors):
            raise ValueError("Each item in 'valve_state_sensors' must be in 'valves' -- cannot " +
                             "place a sensor at a non-existing valve.")

        if not isinstance(pump_state_sensors, list):
            raise TypeError("'pump_state_sensors' must be an instance of 'list[str]' " +
                            f"but not of '{type(pump_state_sensors)}'")
        if any(link not in pumps for link in pump_state_sensors):
            raise ValueError("Each item in 'pump_state_sensors' must be in 'pumps' -- cannot " +
                             "place a sensor at a non-existing pump.")

        if not isinstance(pump_efficiency_sensors, list):
            raise TypeError("'pump_efficiency_sensors' must be an instance of 'list[str]' " +
                            f"but not of '{type(pump_efficiency_sensors)}'")
        if any(link not in pumps for link in pump_efficiency_sensors):
            raise ValueError("Each item in 'pump_efficiency_sensors' must be in 'pumps' -- cannot " +
                             "place a sensor at a non-existing pump.")

        if not isinstance(pump_energyconsumption_sensors, list):
            raise TypeError("'pump_energyconsumption_sensors' must be an instance of 'list[str]' " +
                            f"but not of '{type(pump_energyconsumption_sensors)}'")
        if any(link not in pumps for link in pump_energyconsumption_sensors):
            raise ValueError("Each item in 'pump_energyconsumption_sensors' must be in 'pumps' -- cannot " +
                             "place a sensor at a non-existing pump.")

        if not isinstance(tank_volume_sensors, list):
            raise TypeError("'tank_volume_sensors' must be an instance of 'list[str]' " +
                            f"but not of '{type(tank_volume_sensors)}'")
        if any(n not in tanks for n in tank_volume_sensors):
            raise ValueError("Each item in 'tank_volume_sensors' must be in 'tanks' -- cannot " +
                             "place a sensor at a non-existing tanks.")

        if not isinstance(bulk_species_node_sensors, dict):
            raise TypeError("'bulk_species_node_sensors' must be an instance of 'dict' but not " +
                            f"of '{type(bulk_species_node_sensors)}'")
        if any(bulk_species_id not in bulk_species
               for bulk_species_id in bulk_species_node_sensors.keys()):
            raise ValueError("Unknown bulk species ID in 'bulk_species_node_sensors'")
        if any(node_id not in nodes for node_id in list(itertools.chain(
                *bulk_species_node_sensors.values()))):
            raise ValueError("Unknown node ID in 'bulk_species_node_sensors'")

        if not isinstance(bulk_species_link_sensors, dict):
            raise TypeError("'bulk_species_link_sensors' must be an instance of 'dict' but not " +
                            f"of '{type(bulk_species_link_sensors)}'")
        if any(bulk_species_id not in bulk_species
               for bulk_species_id in bulk_species_link_sensors.keys()):
            raise ValueError("Unknown bulk species ID in 'bulk_species_link_sensors'")
        if any(link_id not in links for link_id in list(itertools.chain(
                *bulk_species_link_sensors.values()))):
            raise ValueError("Unknown link/pipe ID in 'bulk_species_link_sensors'")

        if not isinstance(surface_species_sensors, dict):
            raise TypeError("'surface_species_sensors' must be an instance of 'dict' but not " +
                            f"of '{type(surface_species_sensors)}'")
        if any(surface_species_id not in surface_species_sensors
               for surface_species_id in surface_species_sensors.keys()):
            raise ValueError("Unknown surface species ID in 'surface_species_sensors'")
        if any(link_id not in links for link_id in list(itertools.chain(
                *surface_species_sensors.values()))):
            raise ValueError("Unknown link ID in 'surface_species_sensors'")

        if node_id_to_idx is not None:
            if not isinstance(node_id_to_idx, dict):
                raise TypeError("'node_id_to_idx' must be an instance of 'dict' " +
                                f"but not of '{type(node_id_to_idx)}'")
            if any(n not in nodes for n in node_id_to_idx.keys()):
                raise ValueError("Unknown node ID in 'node_id_to_idx'")

        if link_id_to_idx is not None:
            if not isinstance(link_id_to_idx, dict):
                raise TypeError("'link_id_to_idx' must be an instance of 'dict' " +
                                f"but not of '{type(link_id_to_idx)}'")
            if any(link_id not in links for link_id in link_id_to_idx.keys()):
                raise ValueError("Unknown link/pipe ID in 'link_id_to_idx'")

        if valve_id_to_idx is not None:
            if not isinstance(valve_id_to_idx, dict):
                raise TypeError("'valve_id_to_idx' must be an instance of 'dict' " +
                                f"but not of '{type(valve_id_to_idx)}'")
            if any(v not in valves for v in valve_id_to_idx.keys()):
                raise ValueError("Unknown valve ID in 'valve_id_to_idx'")

        if pump_id_to_idx is not None:
            if not isinstance(pump_id_to_idx, dict):
                raise TypeError("'pump_id_to_idx' must be an instance of 'dict' " +
                                f"but not of '{type(pump_id_to_idx)}'")
            if any(p not in valves for p in pump_id_to_idx.keys()):
                raise ValueError("Unknown pump ID in 'pump_id_to_idx'")

        if tank_id_to_idx is not None:
            if not isinstance(tank_id_to_idx, dict):
                raise TypeError("'tank_id_to_idx' must be an instance of 'dict' " +
                                f"but not of '{type(tank_id_to_idx)}'")
            if any(t not in tanks for t in tank_id_to_idx.keys()):
                raise ValueError("Unknown tank ID in 'tank_id_to_idx'")

        if bulkspecies_id_to_idx is not None:
            if not isinstance(bulkspecies_id_to_idx, dict):
                raise TypeError("'bulkspecies_id_to_idx' must be an instance of 'dict' " +
                                f"but not of '{type(bulkspecies_id_to_idx)}'")
            if any(s not in bulk_species for s in bulkspecies_id_to_idx.keys()):
                raise ValueError("Unknown bulk species ID in 'bulkspecies_id_to_idx'")

        if surfacespecies_id_to_idx is not None:
            if not isinstance(surfacespecies_id_to_idx, dict):
                raise TypeError("'surfacespecies_id_to_idx' must be an instance of 'dict' " +
                                f"but not of '{type(surfacespecies_id_to_idx)}'")
            if any(s not in surface_species for s in surfacespecies_id_to_idx.keys()):
                raise ValueError("Unknown surface species ID in 'surfacespecies_id_to_idx'")

        if not isinstance(flow_unit, int):
            raise TypeError("'flow_unit' must be a an instance of 'int' " +
                            f"but not of '{type(flow_unit)}'")
        if flow_unit not in range(10):
            raise ValueError("Invalid value of 'flow_unit'")

        if quality_unit is not None:
            if not isinstance(quality_unit, int):
                raise TypeError("'quality_mass_unit' must be an instance of 'int' " +
                                f"but not of '{type(quality_unit)}'")
            if quality_unit not in [MASS_UNIT_MG, MASS_UNIT_UG, TIME_UNIT_HRS]:
                raise ValueError("Invalid value of 'quality_unit'")

        if len(bulk_species_mass_unit) != len(bulk_species):
            raise ValueError("Inconsistency between 'bulk_species_mass_unit' and 'bulk_species'")
        if any(not isinstance(mass_unit, int) for mass_unit in bulk_species_mass_unit):
            raise TypeError("All items in 'bulk_species_mass_unit' must be an instance of 'int'")
        if any(mass_unit not in [MASS_UNIT_MG, MASS_UNIT_UG, MASS_UNIT_MOL, MASS_UNIT_MMOL,
                                 MASS_UNIT_CUSTOM]
               for mass_unit in bulk_species_mass_unit):
            raise ValueError("Invalid mass unit in 'bulk_species_mass_unit'")

        if len(surface_species_mass_unit) != len(surface_species):
            raise ValueError("Inconsistency between 'surface_species_mass_unit' " +
                             "and 'surface_species'")
        if any(not isinstance(mass_unit, int) for mass_unit in surface_species_mass_unit):
            raise TypeError("All items in 'surface_species_mass_unit' must be an instance of 'int'")
        if any(mass_unit not in [MASS_UNIT_MG, MASS_UNIT_UG, MASS_UNIT_MOL, MASS_UNIT_MMOL,
                                 MASS_UNIT_CUSTOM]
               for mass_unit in surface_species_mass_unit):
            raise ValueError("Invalid mass unit in 'surface_species_mass_unit'")

        if surface_species_area_unit is not None:
            if not isinstance(surface_species_area_unit, int):
                raise TypeError("'surface_species_area_unit' must be a an instance of 'int' " +
                                f"but not of '{type(surface_species_area_unit)}'")
            if surface_species_area_unit not in [AREA_UNIT_FT2, AREA_UNIT_M2, AREA_UNIT_CM2]:
                raise ValueError("Invalid area unit 'surface_species_area_unit'")

        self.__nodes = nodes
        self.__links = links
        self.__valves = valves
        self.__pumps = pumps
        self.__tanks = tanks
        self.__bulk_species = bulk_species
        self.__surface_species = surface_species
        self.__pressure_sensors = pressure_sensors
        self.__flow_sensors = flow_sensors
        self.__demand_sensors = demand_sensors
        self.__quality_node_sensors = quality_node_sensors
        self.__quality_link_sensors = quality_link_sensors
        self.__valve_state_sensors = valve_state_sensors
        self.__pump_state_sensors = pump_state_sensors
        self.__pump_energyconsumption_sensors = pump_energyconsumption_sensors
        self.__pump_efficiency_sensors = pump_efficiency_sensors
        self.__tank_volume_sensors = tank_volume_sensors
        self.__bulk_species_node_sensors = bulk_species_node_sensors
        self.__bulk_species_link_sensors = bulk_species_link_sensors
        self.__surface_species_sensors = surface_species_sensors
        self.__node_id_to_idx = node_id_to_idx
        self.__link_id_to_idx = link_id_to_idx
        self.__valve_id_to_idx = valve_id_to_idx
        self.__pump_id_to_idx = pump_id_to_idx
        self.__tank_id_to_idx = tank_id_to_idx
        self.__bulkspecies_id_to_idx = bulkspecies_id_to_idx
        self.__surfacespecies_id_to_idx = surfacespecies_id_to_idx
        self.__flow_unit = flow_unit
        self.__quality_unit = quality_unit
        self.__bulk_species_mass_unit = bulk_species_mass_unit
        self.__surface_species_mass_unit = surface_species_mass_unit
        self.__surface_species_area_unit = surface_species_area_unit

        self.__compute_indices()    # Compute indices

        super().__init__(**kwds)

    @staticmethod
    def create_empty_sensor_config(sensor_config):
        """
        Creates an empty sensor configuration from a given sensor configuration
        -- i.e. a clone of the given sensor configuration except that no sensors are set.

        Parameters
        ----------
        sensor_config : :class:`~epyt_flow.simulation.sensor_config.SensorConfig`
            Sensor configuration used as a basis.

        Returns
        -------
        :class:`epyt_flow.simulation.sensor_config.SensorConfig`
            Empty sensor configuration.
        """
        return SensorConfig(nodes=sensor_config.nodes,
                            links=sensor_config.links,
                            valves=sensor_config.valves,
                            pumps=sensor_config.pumps,
                            tanks=sensor_config.tanks,
                            flow_unit=sensor_config.flow_unit,
                            quality_unit=sensor_config.quality_unit,
                            bulk_species=sensor_config.bulk_species,
                            surface_species=sensor_config.surface_species,
                            bulk_species_mass_unit=sensor_config.bulk_species_mass_unit,
                            surface_species_mass_unit=sensor_config.surface_species_mass_unit,
                            surface_species_area_unit=sensor_config.surface_species_area_unit,
                            node_id_to_idx=sensor_config.node_id_to_idx,
                            link_id_to_idx=sensor_config.link_id_to_idx,
                            valve_id_to_idx=sensor_config.valve_id_to_idx,
                            pump_id_to_idx=sensor_config.pump_id_to_idx,
                            tank_id_to_idx=sensor_config.tank_id_to_idx,
                            bulkspecies_id_to_idx=sensor_config.bulkspecies_id_to_idx,
                            surfacespecies_id_to_idx=sensor_config.surfacespecies_id_to_idx)

    def is_empty(self) -> bool:
        """
        Checks if the sensor configuration is empty -- i.e. no sensors are placed.

        Returns
        -------
        `bool`
            True if no sensors are placed, False otherwise.
        """
        if self.__pressure_sensors == [] and self.__flow_sensors == [] \
                and self.__demand_sensors == [] and self.__quality_node_sensors == [] \
                and self.__quality_link_sensors == [] and self.__valve_state_sensors == [] \
                and self.__pump_state_sensors == [] \
                and self.__pump_energyconsumption_sensors == [] \
                and self.__pump_efficiency_sensors == [] and self.__tank_volume_sensors == [] \
                and self.__bulk_species_node_sensors == [] \
                and self.__bulk_species_link_sensors == [] \
                and self.__surface_species_sensors == []:
            return True
        else:
            return False

    def place_sensors_everywhere(self) -> None:
        """
        Places sensors everywhere -- i.e. every possible quantity is monitored
        at every position in the network.
        """
        self.__pressure_sensors = self.__nodes[:]
        self.__demand_sensors = self.__nodes[:]
        self.__flow_sensors = self.__links[:]
        self.__quality_node_sensors = self.__nodes[:]
        self.__quality_link_sensors = self.__links[:]
        self.__pump_state_sensors = self.__pumps[:]
        self.__pump_energyconsumption_sensors = self.__pumps[:]
        self.__pump_efficiency_sensors = self.__pumps[:]
        self.__tank_volume_sensors = self.__tanks[:]
        self.__bulk_species_node_sensors = {species_id: self.__nodes[:]
                                            for species_id in self.__bulk_species}
        self.__bulk_species_link_sensors = {species_id: self.__links[:]
                                            for species_id in self.__bulk_species}
        self.__surface_species_sensors = {species_id: self.__links[:]
                                          for species_id in self.__surface_species}

        self.__compute_indices()

    @property
    def node_id_to_idx(self) -> dict:
        """
        Mapping of a surface node ID to the EPANET index
        (i.e. position in the raw sensor reading data).

        If None, it is assumed that the nodes (in 'nodes') are
        sorted according to their EPANET index.

        Returns
        -------
        `dict`
            Node ID to index mapping.
        """
        return self.__node_id_to_idx

    @property
    def link_id_to_idx(self) -> dict:
        """
        Mapping of a link/pipe ID to the EPANET index
        (i.e. position in the raw sensor reading data).

        If None is given, it is assumed that the links/pipes (in 'links') are
        sorted according to their EPANET index.

        Returns
        -------
        `dict`
            Link/Pipe ID to index mapping.
        """
        return self.__link_id_to_idx

    @property
    def valve_id_to_idx(self) -> dict:
        """
        Mapping of a valve ID to the EPANET index
        (i.e. position in the raw sensor reading data).

        If None, it is assumed that the valves (in 'valves') are
        sorted according to their EPANET index.

        Returns
        -------
        `dict`
            Valve ID to index mapping.
        """
        return self.__valve_id_to_idx

    @property
    def pump_id_to_idx(self) -> dict:
        """
        Mapping of a pump ID to the EPANET index
        (i.e. position in the raw sensor reading data).

        If None, it is assumed that the pumps (in 'pumps') are
        sorted according to their EPANET index.

        Returns
        -------
        `dict`
            Pump ID to index mapping.
        """
        return self.__pump_id_to_idx

    @property
    def tank_id_to_idx(self) -> dict:
        """
        Mapping of a tank ID to the EPANET index
        (i.e. position in the raw sensor reading data).

        If None, it is assumed that the tanks (in 'tanks') are
        sorted according to their EPANET index.

        Returns
        -------
        `dict`
            Tank ID to index mapping.
        """
        return self.__tank_id_to_idx

    @property
    def bulkspecies_id_to_idx(self) -> dict:
        """
        Mapping of a bulk species ID to the EPANET index
        (i.e. position in the raw sensor reading data).

        If None, it is assumed that the bulk species (in 'bulk_species') are
        sorted according to their EPANET index.

        Returns
        -------
        `dict`
            Bulk species ID to index mapping.
        """
        return self.__bulkspecies_id_to_idx

    @property
    def surfacespecies_id_to_idx(self) -> dict:
        """
        Mapping of a surface species ID to the EPANET index
        (i.e. position in the raw sensor reading data).

        If None, it is assumed that the surface species (in 'surface_species') are
        sorted according to their EPANET index.

        Returns
        -------
        `dict`
            Surface species ID to index mapping.
        """
        return self.__surfacespecies_id_to_idx

    def map_node_id_to_idx(self, node_id: str) -> int:
        """
        Gets the index of a given node ID.

        Parameters
        ----------
        node_id : `str`
            Node ID.

        Returns
        -------
        `int`
            Index of the given node.
        """
        if self.__node_id_to_idx is not None:
            return self.__node_id_to_idx[node_id]
        else:
            return self.__nodes.index(node_id)

    def map_link_id_to_idx(self, link_id: str) -> int:
        """
        Gets the index of a given link ID.

        Parameters
        ----------
        link_id : `str`
            Link ID.

        Returns
        -------
        `int`
            Index of the given link.
        """
        if self.__node_id_to_idx is not None:
            return self.__link_id_to_idx[link_id]
        else:
            return self.__links.index(link_id)

    def map_valve_id_to_idx(self, valve_id: str) -> int:
        """
        Gets the index of a given valve ID.

        Parameters
        ----------
        valve_id : `str`
            Valve ID.

        Returns
        -------
        `int`
            Index of the given valve.
        """
        if self.__valve_id_to_idx is not None:
            return self.__valve_id_to_idx[valve_id]
        else:
            return self.__valves.index(valve_id)

    def map_pump_id_to_idx(self, pump_id: str) -> int:
        """
        Gets the index of a given pump ID.

        Parameters
        ----------
        pump_id : `str`
            Pump ID.

        Returns
        -------
        `int`
            Index of the given pump.
        """
        if self.__pump_id_to_idx is not None:
            return self.__pump_id_to_idx[pump_id]
        else:
            return self.__pumps.index(pump_id)

    def map_tank_id_to_idx(self, tank_id: str) -> int:
        """
        Gets the index of a given tank ID.

        Parameters
        ----------
        tank_id : `str`
            Tank ID.

        Returns
        -------
        `int`
            Index of the given tank.
        """
        if self.__tank_id_to_idx is not None:
            return self.__tank_id_to_idx[tank_id]
        else:
            return self.__tanks.index(tank_id)

    def map_bulkspecies_id_to_idx(self, bulk_species_id: str) -> int:
        """
        Gets the index of a given bulk species ID.

        Parameters
        ----------
        bulk_species_id : `str`
            Bulk species ID.

        Returns
        -------
        `int`
            Index of the given bulk species.
        """
        if self.__bulkspecies_id_to_idx is not None:
            return self.__bulkspecies_id_to_idx[bulk_species_id]
        else:
            return self.__bulk_species.index(bulk_species_id)

    def map_surfacespecies_id_to_idx(self, surface_species_id: str) -> int:
        """
        Gets the index of a given surface species ID.

        Parameters
        ----------
        surface_species_id : `str`
            Surface species ID.

        Returns
        -------
        `int`
            Index of the given surface species.
        """
        if self.__surfacespecies_id_to_idx is not None:
            return self.__surfacespecies_id_to_idx[surface_species_id]
        else:
            return self.__surface_species.index(surface_species_id)

    def __compute_indices(self):
        self.__pressure_idx = np.array([self.map_node_id_to_idx(n)
                                        for n in self.__pressure_sensors], dtype=np.int32)
        self.__flow_idx = np.array([self.map_link_id_to_idx(link)
                                    for link in self.__flow_sensors], dtype=np.int32)
        self.__demand_idx = np.array([self.map_node_id_to_idx(n)
                                      for n in self.__demand_sensors], dtype=np.int32)
        self.__quality_node_idx = np.array([self.map_node_id_to_idx(n)
                                            for n in self.__quality_node_sensors], dtype=np.int32)
        self.__quality_link_idx = np.array([self.map_link_id_to_idx(link)
                                            for link in self.__quality_link_sensors],
                                           dtype=np.int32)
        self.__valve_state_idx = np.array([self.map_valve_id_to_idx(v)
                                           for v in self.__valve_state_sensors], dtype=np.int32)
        self.__pump_state_idx = np.array([self.map_pump_id_to_idx(p)
                                          for p in self.__pump_state_sensors], dtype=np.int32)
        self.__pump_efficiency_idx = np.array([self.map_pump_id_to_idx(p)
                                               for p in self.__pump_efficiency_sensors],
                                               dtype=np.int32)
        self.__pump_energyconsumption_idx = np.array([self.map_pump_id_to_idx(p)
                                                      for p in self.__pump_energyconsumption_sensors],
                                                      dtype=np.int32)
        self.__tank_volume_idx = np.array([self.map_tank_id_to_idx(t)
                                           for t in self.__tank_volume_sensors], dtype=np.int32)
        self.__bulk_species_node_idx = np.array([(self.map_bulkspecies_id_to_idx(s),
                                                  [self.map_node_id_to_idx(node_id)
                                                for node_id in self.__bulk_species_node_sensors[s]])
                                                for s in self.__bulk_species_node_sensors.keys()],
                                                dtype=object)
        self.__bulk_species_link_idx = np.array([(self.map_bulkspecies_id_to_idx(s),
                                                  [self.map_link_id_to_idx(link_id)
                                                for link_id in self.__bulk_species_link_sensors[s]])
                                                for s in self.__bulk_species_link_sensors.keys()],
                                                dtype=object)
        self.__surface_species_idx = np.array([(self.map_surfacespecies_id_to_idx(s),
                                                [self.map_link_id_to_idx(link_id)
                                                 for link_id in self.__surface_species_sensors[s]])
                                               for s in self.__surface_species_sensors.keys()],
                                              dtype=object)

        n_pressure_sensors = len(self.__pressure_sensors)
        n_flow_sensors = len(self.__flow_sensors)
        n_demand_sensors = len(self.__demand_sensors)
        n_node_quality_sensors = len(self.__quality_node_sensors)
        n_link_quality_sensors = len(self.__quality_link_sensors)
        n_valve_state_sensors = len(self.__valve_state_sensors)
        n_pump_state_sensors = len(self.__pump_state_sensors)
        n_pump_efficiency_sensors = len(self.__pump_efficiency_sensors)
        n_pump_energyconsumption_sensors = len(self.__pump_energyconsumption_sensors)
        n_tank_volume_sensors = len(self.__tank_volume_sensors)
        n_bulk_species_node_sensors = len(list(itertools.chain(
            *self.__bulk_species_node_sensors.values())))
        n_bulk_species_link_sensors = len(list(itertools.chain(
            *self.__bulk_species_link_sensors.values())))

        pressure_idx_shift = 0
        flow_idx_shift = pressure_idx_shift + n_pressure_sensors
        demand_idx_shift = flow_idx_shift + n_flow_sensors
        node_quality_idx_shift = demand_idx_shift + n_demand_sensors
        link_quality_idx_shift = node_quality_idx_shift + n_node_quality_sensors
        valve_state_idx_shift = link_quality_idx_shift + n_link_quality_sensors
        pump_state_idx_shift = valve_state_idx_shift + n_valve_state_sensors
        pump_efficiency_idx_shift = pump_state_idx_shift + n_pump_state_sensors
        pump_energyconsumption_idx_shift = pump_efficiency_idx_shift + n_pump_efficiency_sensors
        tank_volume_idx_shift = pump_energyconsumption_idx_shift + n_pump_energyconsumption_sensors
        bulk_species_node_idx_shift = tank_volume_idx_shift + n_tank_volume_sensors
        bulk_species_link_idx_shift = bulk_species_node_idx_shift + n_bulk_species_node_sensors
        surface_species_idx_shift = bulk_species_link_idx_shift + n_bulk_species_link_sensors

        def __build_sensors_id_to_idx(sensors: list[str], initial_idx_shift: int) -> dict:
            return {sensor_id: i + initial_idx_shift
                    for sensor_id, i in zip(sensors, range(len(sensors)))}

        def __build_species_sensors_id_to_idx(species_sensors: dict, initial_idx_shift) -> dict:
            r = {}

            cur_idx_shift = initial_idx_shift
            for species_id in species_sensors:
                r[species_id] = {}
                for sensor_id in species_sensors[species_id]:
                    r[species_id][sensor_id] = cur_idx_shift
                    cur_idx_shift += 1

            return r

        mapping = {"pressure": __build_sensors_id_to_idx(self.__pressure_sensors,
                                                         pressure_idx_shift),
                   "flow": __build_sensors_id_to_idx(self.__flow_sensors,flow_idx_shift),
                   "demand": __build_sensors_id_to_idx(self.__demand_sensors,demand_idx_shift),
                   "quality_node": __build_sensors_id_to_idx(self.__quality_node_sensors,
                                                             node_quality_idx_shift),
                   "quality_link": __build_sensors_id_to_idx(self.__quality_link_sensors,
                                                             link_quality_idx_shift),
                   "valve_state": __build_sensors_id_to_idx(self.__valve_state_sensors,
                                                            valve_state_idx_shift),
                   "pump_state": __build_sensors_id_to_idx(self.__pump_state_sensors,
                                                           pump_state_idx_shift),
                   "pump_efficiency": __build_sensors_id_to_idx(self.__pump_efficiency_sensors,
                                                                pump_efficiency_idx_shift),
                   "pump_energyconsumption":
                   __build_sensors_id_to_idx(self.__pump_energyconsumption_sensors,
                                             pump_energyconsumption_idx_shift),
                   "tank_volume": __build_sensors_id_to_idx(self.__tank_volume_sensors,
                                                            tank_volume_idx_shift),
                   "bulk_species_node":
                   __build_species_sensors_id_to_idx(self.__bulk_species_node_sensors,
                                                     bulk_species_node_idx_shift),
                   "bulk_species_link":
                   __build_species_sensors_id_to_idx(self.__bulk_species_link_sensors,
                                                     bulk_species_link_idx_shift),
                   "surface_species":
                   __build_species_sensors_id_to_idx(self.__surface_species_sensors,
                                                     surface_species_idx_shift)}
        self.__sensors_id_to_idx = mapping

    def validate(self, epanet_api: EPyT) -> None:
        """
        Validates this sensor configuration --
        i.e. checks whether all nodes, etc. exist in the .inp file.

        Parameters
        ----------
        epanet_api : `epanet_plus.EPyT <https://epanet-plus.readthedocs.io/en/stable/api.html#epanet_plus.epanet_toolkit.EPyT>`_
            EPANET and EPANET-MSX API.
        """
        if not isinstance(epanet_api, EPyT):
            raise TypeError("'epanet_api' must be an instance of 'epanet-plus.EPyT' " +
                            f"but not of '{type(epanet_api)}'")

        nodes = epanet_api.get_all_nodes_id()
        links = epanet_api.get_all_links_id()
        valves = epanet_api.get_all_valves_id()
        pumps = epanet_api.get_all_pumps_id()
        tanks = epanet_api.get_all_tanks_id()

        bulk_species = []
        surface_species = []
        if epanet_api.msx_file is not None:
            for species_id, species_info in zip(epanet_api.get_all_msx_species_id(),
                                                epanet_api.get_all_msx_species_info()):
                if species_info["type"] == EpanetConstants.MSX_BULK:
                    bulk_species.append(species_id)
                elif species_info["type"] == EpanetConstants.MSX_WALL:
                    surface_species.append(species_id)

        if any(node_id not in nodes for node_id in self.__nodes):
            raise ValueError("Invalid node ID detected -- " +
                             "all given node IDs must exist in the .inp file")
        if any(link_id not in links for link_id in self.__links):
            raise ValueError("Invalid link/pipe ID detected -- all given link/pipe IDs " +
                             "must exist in the .inp file")
        if any(valve_id not in valves for valve_id in self.__valves):
            raise ValueError("Invalid valve ID detected -- all given valve IDs must exist " +
                             "in the .inp file")
        if any(pump_id not in pumps for pump_id in self.__pumps):
            raise ValueError("Invalid pump ID detected -- all given pump IDs must exist " +
                             "in the .inp file")
        if any(tank_id not in tanks for tank_id in self.__tanks):
            raise ValueError("Invalid tank ID detected -- all given tank IDs must exist " +
                             "in the .inp file")
        if any(surface_species_id not in surface_species
               for surface_species_id in self.__surface_species):
            raise ValueError("Invalid surface species ID detected")
        if any(bulk_species_id not in bulk_species for bulk_species_id in self.__bulk_species):
            raise ValueError("Invalid bulk species ID detected")

    @property
    def nodes(self) -> list[str]:
        """
        Gets all node IDs.

        Returns
        -------
        `list[str]`
            All node IDs.
        """
        return self.__nodes.copy()

    @property
    def links(self) -> list[str]:
        """
        Gets all link IDs.

        Returns
        -------
        `list[str]`
            All link IDs.
        """
        return self.__links.copy()

    @property
    def junctions(self) -> list[str]:
        """
        Returns all junction IDs.

        Returns
        -------
        `list[str]`
            All juncitons IDs.
        """
        junctions = self.nodes
        for tank_id in self.tanks:
            junctions.remove(tank_id)

        return junctions

    @property
    def valves(self) -> list[str]:
        """
        Gets all valve IDs (subset of link IDs).

        Returns
        -------
        `list[str]`
            All valve IDs.
        """
        return self.__valves.copy()

    @property
    def pumps(self) -> list[str]:
        """
        Gets all pump IDs (subset of link IDs).

        Returns
        -------
        `list[str]`
            All pump IDs.
        """
        return self.__pumps.copy()

    @property
    def tanks(self) -> list[str]:
        """
        Gets all tank IDs (subset of node IDs).

        Returns
        -------
        `list[str]`
            All tank IDs.
        """
        return self.__tanks.copy()

    @property
    def flow_unit(self) -> int:
        """
        Gets the flow units.
        Note that this specifies all other hydraulic units as well.

        Will be one of the following EPANET constants:

            - EN_CFS = 0 (cubic foot/sec)
            - EN_GPM = 1 (gal/min)
            - EN_MGD = 2 (Million gal/day)
            - EN_IMGD = 3 (Imperial MGD)
            - EN_AFD = 4 (ac-foot/day)
            - EN_LPS = 5 (liter/sec)
            - EN_LPM = 6 (liter/min)
            - EN_MLD = 7 (Megaliter/day)
            - EN_CMH = 8 (cubic meter/hr)
            - EN_CMD = 9 (cubic meter/day)

        Returns
        -------
        `int`
            Flow unit ID.
        """
        return self.__flow_unit

    @property
    def quality_unit(self) -> int:
        """
        Gets the measurement unit ID used in the basic quality analysis.

        Will be one of the following constants:

            - MASS_UNIT_MG = 4      (milligram)
            - MASS_UNIT_UG = 5      (microgram)
            - TIME_UNIT_HRS = 6     (hours)

        Returns
        -------
        `int`
            Mass unit ID.
        """
        return self.__quality_unit

    @property
    def bulk_species(self) -> list[str]:
        """
        Gets all bulk species IDs -- i.e. species that live in the water.

        Returns
        -------
        `list[str]`
            All species IDs.
        """
        return self.__bulk_species.copy()

    @property
    def surface_species(self) -> list[str]:
        """
        Gets all surface species IDs -- i.e. species that live links/pipes.

        Returns
        -------
        `list[str]`
            All species IDs.
        """
        return self.__surface_species.copy()

    @property
    def bulk_species_mass_unit(self) -> list[int]:
        """
        Gets the mass unit of each bulk species.

        Will be one of the following constants:

            - MASS_UNIT_MG = 4      (milligram)
            - MASS_UNIT_UG = 5      (microgram)
            - MASS_UNIT_MOL = 6     (mole)
            - MASS_UNIT_MMOL = 7    (millimole)

        Returns
        -------
        `int`
            Mass unit ID.
        """
        return self.__bulk_species_mass_unit

    @property
    def surface_species_mass_unit(self) -> list[int]:
        """
        Gets the mass unit of each surface species.

        Will be one of the following constants:

            - MASS_UNIT_MG = 4      (milligram)
            - MASS_UNIT_UG = 5      (microgram)
            - MASS_UNIT_MOL = 6     (mole)
            - MASS_UNIT_MMOL = 7    (millimole)

        Returns
        -------
        `int`
            Mass unit ID.
        """
        return self.__surface_species_mass_unit

    @property
    def surface_species_area_unit(self) -> int:
        """
        Gets the surface species area unit.

        Will be one of the following constants:

            - AREA_UNIT_FT2 = 1     (square feet)
            - AREA_UNIT_M2 = 2      (square meters)
            - AREA_UNIT_CM2 = 3     (square centimeters)

        Returns
        -------
        `int`
            Area unit ID.
        """
        return self.__surface_species_area_unit

    @property
    def pressure_sensors(self) -> list[str]:
        """
        Gets all pressure sensors (i.e. IDs of nodes at which a pressure sensor is placed).

        Returns
        -------
        `list[str]`
            All node IDs with a pressure sensor.
        """
        return self.__pressure_sensors.copy()

    @pressure_sensors.setter
    def pressure_sensors(self, pressure_sensors: list[str]) -> None:
        if not isinstance(pressure_sensors, list):
            raise TypeError("'pressure_sensors' must be an instance of 'list[str]' " +
                            f"but not of '{type(pressure_sensors)}'")
        if any(n not in self.__nodes for n in pressure_sensors):
            raise ValueError("Each item in 'pressure_sensors' must be in 'nodes' -- cannot " +
                             "place a sensor at a non-existing node.")

        self.__pressure_sensors = pressure_sensors

        self.__compute_indices()

    @property
    def flow_sensors(self) -> list[str]:
        """
        Gets all flow sensors (i.e. IDs of links at which a flow sensor is placed).

        Returns
        -------
        `list[str]`
            All link IDs with a flow sensor.
        """
        return self.__flow_sensors.copy()

    @flow_sensors.setter
    def flow_sensors(self, flow_sensors: list[str]) -> None:
        if not isinstance(flow_sensors, list):
            raise TypeError("'pressure_sensors' must be an instance of 'list[str]' " +
                            f"but not of '{type(flow_sensors)}'")
        if any(link not in self.__links for link in flow_sensors):
            raise ValueError("Each item in 'flow_sensors' must be in 'links' -- cannot " +
                             "place a sensor at a non-existing link/pipe.")

        self.__flow_sensors = flow_sensors

        self.__compute_indices()

    @property
    def demand_sensors(self) -> list[str]:
        """
        Gets all demand sensors (i.e. IDs of nodes at which a demand sensor is placed).

        Returns
        -------
        `list[str]`
            All node IDs with a demand sensor.
        """
        return self.__demand_sensors.copy()

    @demand_sensors.setter
    def demand_sensors(self, demand_sensors: list[str]) -> None:
        if not isinstance(demand_sensors, list):
            raise TypeError("'demand_sensors' must be an instance of 'list[str]' " +
                            f"but not of '{type(demand_sensors)}'")
        if any(n not in self.__nodes for n in demand_sensors):
            raise ValueError("Each item in 'demand_sensors' must be in 'nodes' -- cannot " +
                             "place a sensor at a non-existing node.")

        self.__demand_sensors = demand_sensors

        self.__compute_indices()

    @property
    def quality_node_sensors(self) -> list[str]:
        """
        Gets all node quality sensors (i.e. IDs of nodes at which a node quality sensor is placed).

        Returns
        -------
        `list[str]`
            All node IDs with a node quality sensor.
        """
        return self.__quality_node_sensors.copy()

    @quality_node_sensors.setter
    def quality_node_sensors(self, quality_node_sensors: list[str]) -> None:
        if not isinstance(quality_node_sensors, list):
            raise TypeError("'quality_node_sensors' must be an instance of 'list[str]' " +
                            f"but not of '{type(quality_node_sensors)}'")
        if any(n not in self.__nodes for n in quality_node_sensors):
            raise ValueError("Each item in 'quality_node_sensors' must be in 'nodes' -- cannot " +
                             "place a sensor at a non-existing node.")

        self.__quality_node_sensors = quality_node_sensors

        self.__compute_indices()

    @property
    def quality_link_sensors(self) -> list[str]:
        """
        Gets all link quality sensors (i.e. IDs of links at which a link quality sensor is placed).

        Returns
        -------
        `list[str]`
            All link IDs with a link quality sensor.
        """
        return self.__quality_link_sensors.copy()

    @quality_link_sensors.setter
    def quality_link_sensors(self, quality_link_sensors: list[str]) -> None:
        if not isinstance(quality_link_sensors, list):
            raise TypeError("'quality_link_sensors' must be an instance of 'list[str]' " +
                            f"but not of '{type(quality_link_sensors)}'")
        if any(link not in self.__links for link in quality_link_sensors):
            raise ValueError("Each item in 'quality_link_sensors' must be in 'links' -- cannot " +
                             "place a sensor at a non-existing link/pipe.")

        self.__quality_link_sensors = quality_link_sensors

        self.__compute_indices()

    @property
    def valve_state_sensors(self) -> list[str]:
        """
        Gets all valve state sensors (i.e. IDs of valves at which a valve state sensor is placed).

        Returns
        -------
        `list[str]`
            All valve IDs with a valve state sensor.
        """
        return self.__valve_state_sensors.copy()

    @valve_state_sensors.setter
    def valve_state_sensors(self, valve_state_sensors: list[str]) -> None:
        if not isinstance(valve_state_sensors, list):
            raise TypeError("'valve_state_sensors' must be an instance of 'list[str]' " +
                            f"but not of '{type(valve_state_sensors)}'")
        if any(link not in self.__valves for link in valve_state_sensors):
            raise ValueError("Each item in 'valve_state_sensors' must be in 'valves' -- cannot " +
                             "place a sensor at a non-existing valves.")

        self.__valve_state_sensors = valve_state_sensors

        self.__compute_indices()

    @property
    def pump_state_sensors(self) -> list[str]:
        """
        Gets all pump state sensors (i.e. IDs of pumps at which a pump state sensor is placed).

        Returns
        -------
        `list[str]`
            All link IDs with a pump state sensor.
        """
        return self.__pump_state_sensors.copy()

    @pump_state_sensors.setter
    def pump_state_sensors(self, pump_state_sensors: list[str]) -> None:
        if not isinstance(pump_state_sensors, list):
            raise TypeError("'pump_state_sensors' must be an instance of 'list[str]' " +
                            f"but not of '{type(pump_state_sensors)}'")
        if any(link not in self.__pumps for link in pump_state_sensors):
            raise ValueError("Each item in 'pump_state_sensors' must be in 'pumps' -- cannot " +
                             "place a sensor at a non-existing pump.")

        self.__pump_state_sensors = pump_state_sensors

        self.__compute_indices()

    @property
    def pump_energyconsumption_sensors(self) -> list[str]:
        """
        Gets all pump energy consumption sensors
        (i.e. IDs of pumps at which the energy consumption is monitored).

        Returns
        -------
        `list[str]`
            All pump IDs with an energy consumption sensor.
        """
        return self.__pump_energyconsumption_sensors.copy()

    @pump_energyconsumption_sensors.setter
    def pump_energyconsumption_sensors(self, pump_energyconsumption_sensors: list[str]) -> None:
        if not isinstance(pump_energyconsumption_sensors, list):
            raise TypeError("'pump_energyconsumption_sensors' must be an instance of 'list[str]' " +
                            f"but not of '{type(pump_energyconsumption_sensors)}'")
        if any(link not in self.__pumps for link in pump_energyconsumption_sensors):
            raise ValueError("Each item in 'pump_energyconsumption_sensors' must be in 'pumps' " +
                             "-- cannot place a sensor at a non-existing pump.")

        self.__pump_energyconsumption_sensors = pump_energyconsumption_sensors

        self.__compute_indices()

    @property
    def pump_efficiency_sensors(self) -> list[str]:
        """
        Gets all pump efficiency sensors
        (i.e. IDs of pumps at which the efficiency is monitored).

        Returns
        -------
        `list[str]`
            All pump IDs with an efficiency sensor.
        """
        return self.__pump_efficiency_sensors.copy()

    @pump_efficiency_sensors.setter
    def pump_efficiency_sensors(self, pump_efficiency_sensors: list[str]) -> None:
        if not isinstance(pump_efficiency_sensors, list):
            raise TypeError("'pump_efficiency_sensors' must be an instance of 'list[str]' " +
                            f"but not of '{type(pump_efficiency_sensors)}'")
        if any(link not in self.__pumps for link in pump_efficiency_sensors):
            raise ValueError("Each item in 'pump_efficiency_sensors' must be in 'pumps' " +
                             "-- cannot place a sensor at a non-existing pump.")

        self.__pump_efficiency_sensors = pump_efficiency_sensors

        self.__compute_indices()

    @property
    def tank_volume_sensors(self) -> list[str]:
        """
        Gets all tank volume sensors (i.e. IDs of tanks at which a tank volume sensor is placed).

        Returns
        -------
        `list[str]`
            All tank IDs with a tank volume sensor.
        """
        return self.__tank_volume_sensors.copy()

    @tank_volume_sensors.setter
    def tank_volume_sensors(self, tank_volume_sensors: list[str]) -> None:
        if not isinstance(tank_volume_sensors, list):
            raise TypeError("'tank_volume_sensors' must be an instance of 'list[str]' " +
                            f"but not of '{type(tank_volume_sensors)}'")
        if any(n not in self.__tanks for n in tank_volume_sensors):
            raise ValueError("Each item in 'tank_volume_sensors' must be in 'tanks' -- cannot " +
                             "place a sensor at a non-existing tanks.")

        self.__tank_volume_sensors = tank_volume_sensors

        self.__compute_indices()

    @property
    def bulk_species_node_sensors(self) -> dict:
        """
        Gets all bulk species node sensors as a dictionary --
        i.e. bulk species IDs as keys and node IDs as values.

        Returns
        -------
        `dict`
            Bulk species sensors -- keys: bulk species IDs, values: node IDs.
        """
        return deepcopy(self.__bulk_species_node_sensors)

    @bulk_species_node_sensors.setter
    def bulk_species_node_sensors(self, bulk_species_sensors: dict) -> None:
        if not isinstance(bulk_species_sensors, dict):
            raise TypeError("'bulk_species_sensors' must be an instance of 'dict' " +
                            f"but not of '{type(bulk_species_sensors)}'")
        if any(species_id not in self.__bulk_species for species_id in bulk_species_sensors.keys()):
            raise ValueError("Unknown bulk species ID in 'bulk_species_sensors'")
        if any(node_id not in self.__nodes for node_id in sum(bulk_species_sensors.values(), [])):
            raise ValueError("Unknown node ID in 'bulk_species_sensors'")

        self.__bulk_species_node_sensors = bulk_species_sensors

        self.__compute_indices()

    @property
    def bulk_species_link_sensors(self) -> dict:
        """
        Gets all bulk species link/pipe sensors as a dictionary --
        i.e. bulk species IDs as keys and link/pipe IDs as values.

        Returns
        -------
        `dict`
            Bulk species sensors -- keys: bulk species IDs, values: link/pipe IDs.
        """
        return deepcopy(self.__bulk_species_link_sensors)

    @bulk_species_link_sensors.setter
    def bulk_species_link_sensors(self, bulk_species_sensors: dict) -> None:
        if not isinstance(bulk_species_sensors, dict):
            raise TypeError("'bulk_species_sensors' must be an instance of 'dict' " +
                            f"but not of '{type(bulk_species_sensors)}'")
        if any(species_id not in self.__bulk_species for species_id in bulk_species_sensors.keys()):
            raise ValueError("Unknown bulk species ID in 'bulk_species_sensors'")
        if any(link_id not in self.__links for link_id in list(itertools.chain(
                *bulk_species_sensors.values()))):
            raise ValueError("Unknown link/pipe ID in 'bulk_species_sensors'")

        self.__bulk_species_link_sensors = bulk_species_sensors

        self.__compute_indices()

    @property
    def surface_species_sensors(self) -> dict:
        """
        Gets all surface species sensors as a dictionary --
        i.e. surface species IDs as keys and link/pipe IDs as values.

        Returns
        -------
        `dict`
            Surface species sensors -- keys: surface species IDs, values: link/pipe IDs.
        """
        return deepcopy(self.__surface_species_sensors)

    @surface_species_sensors.setter
    def surface_species_sensors(self, surface_species_sensors: dict) -> None:
        if not isinstance(surface_species_sensors, dict):
            raise TypeError("'surface_species_sensors' must be an instance of 'dict' " +
                            f"but not of '{type(surface_species_sensors)}'")
        if any(species_id not in self.__surface_species
               for species_id in surface_species_sensors.keys()):
            raise ValueError("Unknown surface species ID in 'surface_species_sensors'")
        if any(link_id not in self.__links
               for link_id in list(itertools.chain(*surface_species_sensors.values()))):
            raise ValueError("Unknown link/pipe ID in 'surface_species_sensors'")

        self.__surface_species_sensors = surface_species_sensors

        self.__compute_indices()

    @property
    def sensors_id_to_idx(self) -> dict:
        """
        Gets a mapping of sensor IDs to indices in the final `Numpy array <https://numpy.org/doc/stable/reference/generated/numpy.ndarray.html>`_ returned by `get_data()`.

        Returns
        -------
        `dict`
            Mapping of sensor IDs to indices in the final `Numpy array <https://numpy.org/doc/stable/reference/generated/numpy.ndarray.html>`_.
        """
        return deepcopy(self.__sensors_id_to_idx)

    def get_as_dict(self) -> dict:
        """
        Gets the sensor configuration as a dictionary.

        Returns
        -------
        `dict`
            Dictionary of set sensors -- the keys are the sensor types.
        """
        r = {}

        if self.__pressure_sensors != []:
            r["pressure"] = self.__pressure_sensors
        if self.__flow_sensors != []:
            r["flow"] = self.__flow_sensors
        if self.__demand_sensors != []:
            r["demand"] = self.__demand_sensors
        if self.__tank_volume_sensors != []:
            r["tank_volume"] = self.__tank_volume_sensors
        if self.__valve_state_sensors != []:
            r["valve_state"] = self.__valve_state_sensors
        if self.__pump_state_sensors != []:
            r["pump_state"] = self.__pump_state_sensors
        if self.__pump_efficiency_sensors != []:
            r["pump_efficiency"] = self.__pump_efficiency_sensors
        if self.__pump_energyconsumption_sensors != []:
            r["pump_energyconsumption"] = self.__pump_energyconsumption_sensors
        if self.__quality_node_sensors != []:
            r["node_quality"] = self.__quality_node_sensors
        if self.__quality_link_sensors != []:
            r["link_quality"] = self.__quality_link_sensors
        if self.__bulk_species_node_sensors != {}:
            r["node_bulk_species"] = self.__bulk_species_node_sensors
        if self.__bulk_species_link_sensors != {}:
            r["link_bulk_species"] = self.__bulk_species_link_sensors
        if self.__surface_species_sensors != {}:
            r["surface_species"] = self.__surface_species_sensors

        return r

    def get_attributes(self) -> dict:
        attr = {"nodes": self.__nodes, "links": self.__links,
                "valves": self.__valves, "pumps": self.__pumps,
                "tanks": self.__tanks, "bulk_species": self.__bulk_species,
                "surface_species": self.__surface_species,
                "pressure_sensors": self.__pressure_sensors,
                "flow_sensors": self.__flow_sensors,
                "demand_sensors": self.__demand_sensors,
                "quality_node_sensors": self.__quality_node_sensors,
                "quality_link_sensors": self.__quality_link_sensors,
                "valve_state_sensors": self.__valve_state_sensors,
                "pump_state_sensors": self.__pump_state_sensors,
                "pump_efficiency_sensors": self.__pump_efficiency_sensors,
                "pump_energyconsumption_sensors": self.__pump_energyconsumption_sensors,
                "tank_volume_sensors": self.__tank_volume_sensors,
                "bulk_species_node_sensors": self.__bulk_species_node_sensors,
                "bulk_species_link_sensors": self.__bulk_species_link_sensors,
                "surface_species_sensors": self.__surface_species_sensors,
                "node_id_to_idx": self.__node_id_to_idx,
                "link_id_to_idx": self.__link_id_to_idx,
                "valve_id_to_idx": self.__valve_id_to_idx,
                "pump_id_to_idx": self.__pump_id_to_idx,
                "tank_id_to_idx": self.__tank_id_to_idx,
                "bulkspecies_id_to_idx": self.__bulkspecies_id_to_idx,
                "surfacespecies_id_to_idx": self.__surfacespecies_id_to_idx,
                "flow_unit": self.__flow_unit,
                "quality_unit": self.__quality_unit,
                "bulk_species_mass_unit": self.__bulk_species_mass_unit,
                "surface_species_mass_unit": self.__surface_species_mass_unit,
                "surface_species_area_unit": self.__surface_species_area_unit}

        return super().get_attributes() | attr

    def __eq__(self, other) -> bool:
        if not isinstance(other, SensorConfig):
            raise TypeError("Can not compare 'SensorConfig' instance " +
                            f"with '{type(other)}' instance")

        return self.__nodes == other.nodes and self.__links == other.links \
            and self.__valves == other.valves and self.__pumps == other.pumps \
            and self.__tanks == other.tanks and self.__bulk_species == other.bulk_species \
            and self.__surface_species == other.surface_species \
            and self.__pressure_sensors == other.pressure_sensors \
            and self.__flow_sensors == other.flow_sensors \
            and self.__demand_sensors == other.demand_sensors \
            and self.__quality_node_sensors == other.quality_node_sensors \
            and self.__quality_link_sensors == other.quality_link_sensors \
            and self.__valve_state_sensors == other.valve_state_sensors \
            and self.__pump_state_sensors == other.pump_state_sensors \
            and self.__pump_efficiency_sensors == other.pump_efficiency_sensors \
            and self.__pump_energyconsumption_sensors == other.pump_energyconsumption_sensors \
            and self.__tank_volume_sensors == other.tank_volume_sensors \
            and self.__bulk_species_node_sensors == other.bulk_species_node_sensors \
            and self.__bulk_species_link_sensors == other.bulk_species_link_sensors \
            and self.__surface_species_sensors == other.surface_species_sensors \
            and self.__flow_unit == other.flow_unit \
            and self.__quality_unit == other.quality_unit \
            and self.__bulk_species_mass_unit == other.bulk_species_mass_unit \
            and self.__surface_species_mass_unit == other.surface_species_mass_unit \
            and self.__surface_species_area_unit == other.surface_species_area_unit \
            and self.__node_id_to_idx == other.node_id_to_idx \
            and self.__link_id_to_idx == other.link_id_to_idx \
            and self.__valve_id_to_idx == other.valve_id_to_idx \
            and self.__pump_id_to_idx == other.pump_id_to_idx \
            and self.__tank_id_to_idx == other.tank_id_to_idx \
            and self.__bulkspecies_id_to_idx == other.bulkspecies_id_to_idx \
            and self.__surfacespecies_id_to_idx == other.surfacespecies_id_to_idx

    def __str__(self) -> str:
        return f"nodes: {self.__nodes} links: {self.__links} valves: {self.__valves} " +\
            f"pumps: {self.__pumps} tanks: {self.__tanks} bulk_species: {self.__bulk_species} " +\
            f"surface_species: {self.__surface_species} " + \
            f"node_id_to_idx: {self.__node_id_to_idx} link_id_to_idx: {self.__link_id_to_idx} " +\
            f"pump_id_to_idx: {self.__pump_id_to_idx} tank_id_to_idx: {self.__tank_id_to_idx} " +\
            f"valve_id_to_idx: {self.__valve_id_to_idx} " +\
            f"bulkspecies_id_to_idx: {self.__bulkspecies_id_to_idx} " +\
            f"surfacespecies_id_to_idx: {self.__surfacespecies_id_to_idx} " +\
            f"pressure_sensors: {self.__pressure_sensors} flow_sensors: {self.__flow_sensors} " +\
            f"demand_sensors: {self.__demand_sensors} " +\
            f"quality_node_sensors: {self.__quality_node_sensors} " +\
            f"quality_link_sensors: {self.__quality_link_sensors} " +\
            f"valve_state_sensors: {self.__valve_state_sensors} " +\
            f"pump_state_sensors: {self.__pump_state_sensors} " +\
            f"pump_efficiency_sensors: {self.__pump_efficiency_sensors} " +\
            f"pump_energyconsumption_sensors: {self.__pump_energyconsumption_sensors} " +\
            f"tank_volume_sensors: {self.__tank_volume_sensors} " +\
            f"bulk_species_node_sensors: {self.__bulk_species_node_sensors} " +\
            f"bulk_species_link_sensors: {self.__bulk_species_link_sensors} " +\
            f"surface_species_sensors: {self.__surface_species_sensors} " +\
            f"flow_unit: {flowunit_to_str(self.__flow_unit)} " +\
            f"quality_unit: {qualityunit_to_str(self.__quality_unit)} " +\
            "bulk_species_mass_unit: " +\
            f"{list(map(massunit_to_str, self.__bulk_species_mass_unit))} " +\
            "surface_species_mass_unit: " +\
            f"{list(map(massunit_to_str, self.__surface_species_mass_unit))} " +\
            f"surface_species_area_unit: {areaunit_to_str(self.__surface_species_area_unit)}"

    def get_bulk_species_mass_unit_id(self, bulk_species_id: str) -> int:
        """
        Returns the mass unit of a given bulk species.

        Parameters
        ----------
        bulk_species_id : `str`
            ID of the bulk species.

        Returns
        -------
        `int`
            ID of the mass unit.

            Will be one of the following constant:

                - MASS_UNIT_MG   = 4
                - MASS_UNIT_UG   = 5
                - MASS_UNIT_MOL  = 6
                - MASS_UNIT_MMOL = 7
        """
        return self.__bulk_species_mass_unit[self.map_bulkspecies_id_to_idx(bulk_species_id)]

    def get_surface_species_mass_unit_id(self, surface_species_id: str) -> int:
        """
        Returns the mass unit of a given surface species.

        Parameters
        ----------
        surface_species_id : `str`
            ID of the surface species.

        Returns
        -------
        `int`
            ID of the mass unit.

            Will be one of the following constant:

                - MASS_UNIT_MG   = 4
                - MASS_UNIT_UG   = 5
                - MASS_UNIT_MOL  = 6
                - MASS_UNIT_MMOL = 7
        """
        return self.__surface_species_mass_unit[self.map_surfacespecies_id_to_idx(
            surface_species_id)]

    def compute_readings(self, pressures: np.ndarray, flows: np.ndarray, demands: np.ndarray,
                         nodes_quality: np.ndarray, links_quality: np.ndarray,
                         pumps_state: np.ndarray, pumps_efficiency: np.ndarray,
                         pumps_energyconsumption: np.ndarray, valves_state: np.ndarray,
                         tanks_volume: np.ndarray, bulk_species_node_concentrations: np.ndarray,
                         bulk_species_link_concentrations: np.ndarray,
                         surface_species_concentrations: np.ndarray) -> np.ndarray:
        """
        Applies the sensor configuration to a set of raw simulation results -- i.e. computes
        the sensor readings as an array.

        Columns (i.e. sensor readings) are ordered as follows:

            1. Pressures
            2. Flows
            3. Demands
            4. Nodes quality
            5. Links quality
            6. Valve state
            7. Pumps state
            8. Pumps efficiency
            9. Pumps energy consumption
            10. Tanks volume
            11. Surface species concentrations
            12. Bulk species nodes concentrations
            13. Bulk species links concentrations

        Parameters
        ----------
        pressures : `numpy.ndarray <https://numpy.org/doc/stable/reference/generated/numpy.ndarray.html>`_
            Pressure values at all nodes.
        flows : `numpy.ndarray <https://numpy.org/doc/stable/reference/generated/numpy.ndarray.html>`_
            Flow values at all links/pipes.
        demands : `numpy.ndarray <https://numpy.org/doc/stable/reference/generated/numpy.ndarray.html>`_
            Demand values at all nodes.
        nodes_quality : `numpy.ndarray <https://numpy.org/doc/stable/reference/generated/numpy.ndarray.html>`_
            Quality values at all nodes.
        links_quality : `numpy.ndarray <https://numpy.org/doc/stable/reference/generated/numpy.ndarray.html>`_
            Quality values at all links/pipes.
        pumps_state : `numpy.ndarray <https://numpy.org/doc/stable/reference/generated/numpy.ndarray.html>`_
            States of all pumps.
        pumps_efficiency : `numpy.ndarray <https://numpy.org/doc/stable/reference/generated/numpy.ndarray.html>`_
            Efficiency of all pumps.
        pumps_energyconsumption : `numpy.ndarray <https://numpy.org/doc/stable/reference/generated/numpy.ndarray.html>`_
            Energy consumption of all pumps.
        valves_state : `numpy.ndarray <https://numpy.org/doc/stable/reference/generated/numpy.ndarray.html>`_
            States of all valves.
        tanks_volume : `numpy.ndarray <https://numpy.org/doc/stable/reference/generated/numpy.ndarray.html>`_
            Water volume in all tanks.
        bulk_species_node_concentrations : `numpy.ndarray <https://numpy.org/doc/stable/reference/generated/numpy.ndarray.html>`_
            Bulk species concentrations at all nodes.

            Expect a three-dimensional array: First dimension denotes time,
            second dimension corresponds to species ID,
            and third dimension contains the concentration.
        bulk_species_link_concentrations : `numpy.ndarray <https://numpy.org/doc/stable/reference/generated/numpy.ndarray.html>`_
            Bulk species concentrations at all links/pipes.

            Expect a three-dimensional array: First dimension denotes time,
            second dimension corresponds to species ID,
            and third dimension contains the concentration.
        surface_species_concentrations : `numpy.ndarray <https://numpy.org/doc/stable/reference/generated/numpy.ndarray.html>`_
            Surface species concentrations at all links/pipes.

            Expect a three-dimensional array: First dimension denotes time,
            second dimension corresponds to species ID,
            and third dimension contains the concentration.

        Returns
        -------
        `numpy.ndarray <https://numpy.org/doc/stable/reference/generated/numpy.ndarray.html>`_
            Sensor readings.
        """
        data = []

        if pressures is not None:
            data.append(pressures[:, self.__pressure_idx])
        else:
            if len(self.__pressure_sensors) != 0:
                raise ValueError("Pressure readings requested but no pressure data is given")

        if flows is not None:
            data.append(flows[:, self.__flow_idx])
        else:
            if len(self.__flow_sensors) != 0:
                raise ValueError("Flow readings requested but no flow data is given")

        if demands is not None:
            data.append(demands[:, self.__demand_idx])
        else:
            if len(self.__demand_sensors) != 0:
                raise ValueError("Demand readings requested but no demand data is given")

        if nodes_quality is not None:
            data.append(nodes_quality[:, self.__quality_node_idx])
        else:
            if len(self.__quality_node_sensors) != 0:
                raise ValueError("Node water quality readings requested " +
                                 "but no water quality data at nodes is given")

        if links_quality is not None:
            data.append(links_quality[:, self.__quality_link_idx])
        else:
            if len(self.__quality_link_sensors) != 0:
                raise ValueError("Link/Pipe water quality readings requested " +
                                 "but no water quality data at links/pipes is given")

        if valves_state is not None:
            data.append(valves_state[:, self.__valve_state_idx])
        else:
            if len(self.__valve_state_sensors) != 0:
                raise ValueError("Valve states readings requested " +
                                 "but no valve state data is given")

        if pumps_state is not None:
            data.append(pumps_state[:, self.__pump_state_idx])
        else:
            if len(self.__pump_state_sensors) != 0:
                raise ValueError("Pump states readings requested " +
                                 "but no pump state data is given")

        if pumps_efficiency is not None:
            data.append(pumps_efficiency[:, self.__pump_efficiency_idx])
        else:
            if len(self.__pump_efficiency_sensors) != 0:
                raise ValueError("Pump efficiency readings requested " +
                                 "but no pump efficiency data is given")

        if pumps_energyconsumption is not None:
            data.append(pumps_energyconsumption[:, self.__pump_energyconsumption_idx])
        else:
            if len(self.__pump_energyconsumption_sensors) != 0:
                raise ValueError("Pump energy consumption readings requested " +
                                 "but no pump energy consumption data is given")

        if tanks_volume is not None:
            data.append(tanks_volume[:, self.__tank_volume_idx])
        else:
            if len(self.__tank_volume_sensors) != 0:
                raise ValueError("Water volumes in tanks is requested but no " +
                                 "tank water volume data is given")

        if surface_species_concentrations is not None:
            for species_idx, links_idx in self.__surface_species_idx:
                data.append(surface_species_concentrations[:, species_idx, links_idx].
                            reshape(-1, len(links_idx)))
        else:
            if len(self.__surface_species_sensors) != 0:
                raise ValueError("Surface species concentratinons requested but no " +
                                 "surface species concentration data is given")

        if bulk_species_node_concentrations is not None:
            for species_idx, nodes_idx in self.__bulk_species_node_idx:
                data.append(bulk_species_node_concentrations[:, species_idx, nodes_idx].
                            reshape(-1, len(nodes_idx)))
        else:
            if len(self.__bulk_species_node_sensors) != 0:
                raise ValueError("Bulk species concentratinons requested but no " +
                                 "bulk species node concentration data is given")

        if bulk_species_link_concentrations is not None:
            for species_idx, links_idx in self.__bulk_species_link_idx:
                data.append(bulk_species_link_concentrations[:, species_idx, links_idx].
                            reshape(-1, len(links_idx)))
        else:
            if len(self.__bulk_species_link_sensors) != 0:
                raise ValueError("Bulk species concentratinons requested but no " +
                                 "bulk species link/pipe concentration data is given")

        return np.concatenate(data, axis=1)

    def get_index_of_reading(self, pressure_sensor: str = None, flow_sensor: str = None,
                             demand_sensor: str = None, node_quality_sensor: str = None,
                             link_quality_sensor: str = None, valve_state_sensor: str = None,
                             pump_state_sensor: str = None, pump_efficiency_sensor: str = None,
                             pump_energyconsumption_sensor: str = None,
                             tank_volume_sensor: str = None,
                             bulk_species_node_sensor: tuple[str, str] = None,
                             bulk_species_link_sensor: tuple[str, str] = None,
                             surface_species_sensor: tuple[str, str] = None) -> int:
        """
        Gets the index of a particular sensor in the final sensor readings array.

        Note that only one sensor ID is converted to an index. In case of multiple sensor IDs,
        call this function for each sensor ID separately.

        .. note::

            This function only returns the correct results if the sensor configuraton is NOT frozen!

        Parameters
        ----------
        pressure_sensor : `str`
            ID of the pressure sensor.
        flow_sensor : `str`
            ID of the flow sensor.
        demand_sensor : `str`
            ID of the demand sensor.
        node_quality_sensor : `str`
            ID of the quality sensor (at a node).
        link_quality_sensor : `str`
            ID of the quality sensor (at a link/pipe).
        valve_state_sensor : `str`
            ID of the state sensor (at a valve).
        pump_state_sensor : `str`
            ID of the state sensor (at a pump).
        pump_efficiency_sensor : `str`
            ID of the efficiency sensor (at a pump).
        pump_energyconsumption_sensor : `str`
            ID of the energy consumption sensor (at a pump).
        tank_volume_sensor : `str`
            ID of the water volume sensor (at a tank)
        bulk_species_node_sensor : `tuple[str, str]`
            Tuple of bulk species ID and sensor node ID.
        bulk_species_link_sensor : `tuple[str, str]`
            Tuple of bulk species ID and sensor link/pipe ID.
        surface_species_sensor : `tuple[str, str]`
            Tuple of surface species ID and sensor link/pipe ID.
        """
        if pressure_sensor is not None:
            return self.__sensors_id_to_idx["pressure"][pressure_sensor]
        elif flow_sensor is not None:
            return self.__sensors_id_to_idx["flow"][flow_sensor]
        elif demand_sensor is not None:
            return self.__sensors_id_to_idx["demand"][demand_sensor]
        elif node_quality_sensor is not None:
            return self.__sensors_id_to_idx["quality_node"][node_quality_sensor]
        elif link_quality_sensor is not None:
            return self.__sensors_id_to_idx["quality_link"][link_quality_sensor]
        elif valve_state_sensor is not None:
            return self.__sensors_id_to_idx["valve_state"][valve_state_sensor]
        elif pump_state_sensor is not None:
            return self.__sensors_id_to_idx["pump_state"][pump_state_sensor]
        elif pump_efficiency_sensor is not None:
            return self.__sensors_id_to_idx["pump_efficiency"][pump_efficiency_sensor]
        elif pump_energyconsumption_sensor is not None:
            return self.__sensors_id_to_idx["pump_energyconsumption"][pump_energyconsumption_sensor]
        elif tank_volume_sensor is not None:
            return self.__sensors_id_to_idx["tank_volume"][tank_volume_sensor]
        elif surface_species_sensor is not None:
            species_id, sensor_id = surface_species_sensor
            return self.__sensors_id_to_idx["surface_species"][species_id][sensor_id]
        elif bulk_species_node_sensor is not None:
            species_id, sensor_id = bulk_species_node_sensor
            return self.__sensors_id_to_idx["bulk_species_node"][species_id][sensor_id]
        elif bulk_species_link_sensor is not None:
            species_id, sensor_id = bulk_species_link_sensor
            return self.__sensors_id_to_idx["bulk_species_link"][species_id][sensor_id]
        else:
            raise ValueError("No sensor given")



"""
Module provides a base class for sensor reading events such as sensor faults.
"""
from abc import abstractmethod
import warnings
import numpy

from .event import Event
from ..sensor_config import SensorConfig, SENSOR_TYPE_NODE_PRESSURE, SENSOR_TYPE_NODE_QUALITY, \
    SENSOR_TYPE_NODE_DEMAND, SENSOR_TYPE_LINK_FLOW, SENSOR_TYPE_LINK_QUALITY, \
    SENSOR_TYPE_VALVE_STATE, SENSOR_TYPE_PUMP_STATE, SENSOR_TYPE_TANK_VOLUME, \
    SENSOR_TYPE_NODE_BULK_SPECIES, SENSOR_TYPE_LINK_BULK_SPECIES, SENSOR_TYPE_SURFACE_SPECIES


class SensorReadingEvent(Event):
    """
    Base class for a sensor reading event -- i.e. an event directly affecting sensor readings.

    Parameters
    ----------
    sensor_id : `str`
        ID of the sensor that is affected by this event.
    sensor_type : `int`
        Type of the sensor that is specified in 'sensor_id'.
        Must be one of the following pre-defined constants:

        - SENSOR_TYPE_NODE_PRESSURE     = 1
        - SENSOR_TYPE_NODE_QUALITY      = 2
        - SENSOR_TYPE_NODE_DEMAND       = 3
        - SENSOR_TYPE_LINK_FLOW         = 4
        - SENSOR_TYPE_LINK_QUALITY      = 5
        - SENSOR_TYPE_VALVE_STATE       = 6
        - SENSOR_TYPE_PUMP_STATE        = 7
        - SENSOR_TYPE_TANK_VOLUME       = 8
        - SENSOR_TYPE_NODE_BULK_SPECIES = 9
        - SENSOR_TYPE_NODE_LINK_SPECIES = 10
        - SENSOR_TYPE_SURFACE_SPECIES   = 11
    """
    def __init__(self, sensor_id: str, sensor_type: int, **kwds):
        if not isinstance(sensor_id, str):
            raise TypeError("'sensor_id' must be an instance of 'str' but not of " +
                            f"'{type(sensor_id)}'")
        if not isinstance(sensor_type, int):
            raise TypeError("'sensor_type' mut be an instance of 'int' but not of " +
                            f"'{type(sensor_type)}'")
        if not 1 <= sensor_type <= 10:
            raise ValueError("Invalid value of 'sensor_type'")

        self.__sensor_id = sensor_id
        self.__sensor_type = sensor_type

        super().__init__(**kwds)

    def validate(self, sensor_config: SensorConfig) -> None:
        """
        Validates this sensor reading event -- i.e. checks whether the affected
        sensor is part of the given sensor configuration.

        Parameters
        ----------
        sensor_config : :class:`~epyt_flow.simulation.sensor_config.SensorConfig`
            Sensor configuration.
        """
        if not isinstance(sensor_config, SensorConfig):
            raise TypeError("'sensor_config' must be an instance of " +
                            "'epyt_flow.simulation.SensorConfig' but not of " +
                            f"'{type(sensor_config)}'")

        def __show_warning() -> None:
            warnings.warn("Event does not have any effect because there is " +
                          f"no sensor at '{self.__sensor_id}'")

        if self.__sensor_type == SENSOR_TYPE_NODE_PRESSURE:
            if self.__sensor_id not in sensor_config.pressure_sensors:
                __show_warning()
        elif self.__sensor_type == SENSOR_TYPE_NODE_QUALITY:
            if self.__sensor_id not in sensor_config.quality_node_sensors:
                __show_warning()
        elif self.__sensor_type == SENSOR_TYPE_NODE_DEMAND:
            if self.__sensor_id not in sensor_config.demand_sensors:
                __show_warning()
        elif self.__sensor_type == SENSOR_TYPE_LINK_FLOW:
            if self.__sensor_id not in sensor_config.flow_sensors:
                __show_warning()
        elif self.__sensor_type == SENSOR_TYPE_LINK_QUALITY:
            if self.__sensor_id not in sensor_config.quality_link_sensors:
                __show_warning()
        elif self.__sensor_type == SENSOR_TYPE_VALVE_STATE:
            if self.__sensor_id not in sensor_config.valve_state_sensors:
                __show_warning()
        elif self.__sensor_type == SENSOR_TYPE_PUMP_STATE:
            if self.__sensor_id not in sensor_config.pump_state_sensors:
                __show_warning()
        elif self.__sensor_type == SENSOR_TYPE_TANK_VOLUME:
            if self.__sensor_id not in sensor_config.tank_volume_sensors:
                __show_warning()
        elif self.__sensor_type == SENSOR_TYPE_NODE_BULK_SPECIES:
            sensor_present = False
            for _, sensors_id in sensor_config.bulk_species_node_sensors.items():
                if self.__sensor_id in sensors_id:
                    sensor_present = True
                    break
            if sensor_present is False:
                __show_warning()
        elif self.__sensor_type == SENSOR_TYPE_LINK_BULK_SPECIES:
            sensor_present = False
            for _, sensors_id in sensor_config.bulk_species_link_sensors.items():
                if self.__sensor_id in sensors_id:
                    sensor_present = True
                    break
            if sensor_present is False:
                __show_warning()
        elif self.__sensor_type == SENSOR_TYPE_SURFACE_SPECIES:
            sensor_present = False
            for _, sensors_id in sensor_config.surface_species_sensors.items():
                if self.__sensor_id in sensors_id:
                    sensor_present = True
                    break
            if sensor_present is False:
                __show_warning()

    @property
    def sensor_id(self) -> str:
        """
        Gets the ID of the node or link that is affected by this event.

        Returns
        -------
        `str`
            Node or link ID.
        """
        return self.__sensor_id

    @property
    def sensor_type(self) -> int:
        """
        Gets the sensor type code.

        Returns
        -------
        `int`
            Sensor type code.
        """
        return self.__sensor_type

    def get_attributes(self) -> dict:
        return super().get_attributes() | {"sensor_id": self.__sensor_id,
                                           "sensor_type": self.__sensor_type}

    def __eq__(self, other) -> bool:
        if not isinstance(other, SensorReadingEvent):
            raise TypeError("Can not compare 'SensorReadingEvent' instance " +
                            f"with '{type(other)}' instance")

        return super().__eq__(other) and self.__sensor_id == other.sensor_id \
            and self.__sensor_type == other.sensor_type

    def __str__(self) -> str:
        return f"{super().__str__()} sensor_id: {self.__sensor_id} " +\
            f"sensor_type: {self.__sensor_type}"

    def __call__(self, sensor_readings: numpy.ndarray,
                 sensor_readings_time: numpy.ndarray) -> numpy.ndarray:
        return self.apply(sensor_readings, sensor_readings_time)

    @abstractmethod
    def apply(self, sensor_readings: numpy.ndarray,
              sensor_readings_time: numpy.ndarray) -> numpy.ndarray:
        """
        Applies the sensor reading event to sensor reading values -- i.e.
        modify the sensor readings.

        Parameters
        ----------
        sensor_readings : `numpy.ndarray <https://numpy.org/doc/stable/reference/generated/numpy.ndarray.html>`_
            Original sensor readings.
        sensor_readings_time : `numpy.ndarray <https://numpy.org/doc/stable/reference/generated/numpy.ndarray.html>`_
            Time (seconds since simulation start) for each sensor reading row in 'sensor_readings'.

        Returns
        -------
        `numpy.ndarray <https://numpy.org/doc/stable/reference/generated/numpy.ndarray.html>`_
            Modified sensor readings.
        """
        raise NotImplementedError()



"""
Module provides a class for visualizing scenarios.
"""
from typing import Optional, Union, List, Tuple

import numpy as np

import matplotlib.pyplot as plt
from matplotlib.animation import FuncAnimation
import matplotlib as mpl
import networkx.drawing.nx_pylab as nxp
from svgpath2mpl import parse_path

from ..simulation.scenario_simulator import ScenarioSimulator
from ..simulation.scada.scada_data import ScadaData
from ..visualization import JunctionObject, EdgeObject, ColorScheme, \
    epyt_flow_colors

PUMP_PATH = ('M 202.5 93 A 41.5 42 0 0 0 161 135 A 41.5 42 0 0 0 202.5 177 A '
             '41.5 42 0 0 0 244 135 A 41.5 42 0 0 0 241.94922 122 L 278 122 '
             'L 278 93 L 203 93 L 203 93.011719 A 41.5 42 0 0 0 202.5 93 z')
RESERVOIR_PATH = ('M 325 41 A 43 24.5 0 0 0 282.05664 65 L 282 65 L 282 65.5 '
                  'L 282 163 L 282 168 L 282 216 L 305 216 L 305 168 L 345 '
                  '168 L 345 216 L 368 216 L 368 168 L 368 163 L 368 65.5 L '
                  '368 65 L 367.98047 65 A 43 24.5 0 0 0 325 41 z')
TANK_PATH = ('M 325 41 A 43 24.5 0 0 0 282.05664 65 L 282 65 L 282 65.5 L 282 '
             '185 L 368 185 L 368 65.5 L 368 65 L 367.98047 65 A 43 24.5 0 0'
             ' 0 325 41 z')
VALVE_PATH = ('M 9.9999064 9.9999064 L 9.9999064 110 L 69.999862 59.999955 L '
              '9.9999064 9.9999064 z M 69.999862 59.999955 L 129.99982 110 L '
              '129.99982 9.9999064 L 69.999862 59.999955 z')


class Marker:
    """
    The Marker class provides svg representations of hydraulic components
    (pump, reservoir, tank and valve), which are loaded from their respective
    svg paths and transformed into
    `~matplotlib.path.Path <https://matplotlib.org/stable/api/path_api.html#matplotlib.path.Path>`_
    objects in order to be used with the matplotlib library.

    Attributes
    ----------
    pump : `matplotlib.path.Path <https://matplotlib.org/stable/api/path_api.html#matplotlib.path.Path>`_
        Marker for the pump, loaded from PUMP_PATH.
    reservoir : `matplotlib.path.Path <https://matplotlib.org/stable/api/path_api.html#matplotlib.path.Path>`_
        Marker for the reservoir, loaded from RESERVOIR_PATH.
    tank : `matplotlib.path.Path <https://matplotlib.org/stable/api/path_api.html#matplotlib.path.Path>`_
        Marker for the tank, loaded from TANK_PATH.
    valve : `matplotlib.path.Path <https://matplotlib.org/stable/api/path_api.html#matplotlib.path.Path>`_
        Marker for the valve, loaded from VALVE_PATH.

    Methods
    -------
    __marker_from_path(path, scale_p=1)
        Loads and applies transformations to the marker shape from the given
        path.
    """

    def __init__(self):
        """
        Initializes the Marker class and assigns
        `matplotlib.path.Path <https://matplotlib.org/stable/api/path_api.html#matplotlib.path.Path>`_
        markers for pump, reservoir, tank, and valve components.
        """
        self.pump = self.__marker_from_path(PUMP_PATH, 2)
        self.reservoir = self.__marker_from_path(RESERVOIR_PATH)
        self.tank = self.__marker_from_path(TANK_PATH)
        self.valve = self.__marker_from_path(VALVE_PATH)

    @staticmethod
    def __marker_from_path(path: str, scale_p: int = 1) -> mpl.path.Path:
        """
        Loads the marker from the specified path and adjusts it representation
        by aligning, rotating and scaling it.

        Parameters
        ----------
        path : `str`
            The svg path describing the marker shape.
        scale_p : `float`, optional
            Scaling factor for the marker (default is 1).

        Returns
        -------
        `matplotlib.path.Path <https://matplotlib.org/stable/api/path_api.html#matplotlib.path.Path>`_
            The transformed marker object after loading and adjusting it.
        """
        marker_tmp = parse_path(path)
        marker_tmp.vertices -= marker_tmp.vertices.mean(axis=0)
        marker_tmp = marker_tmp.transformed(
            mpl.transforms.Affine2D().rotate_deg(180))
        marker_tmp = marker_tmp.transformed(
            mpl.transforms.Affine2D().scale(-scale_p, scale_p))
        return marker_tmp


class ScenarioVisualizer:
    """
    This class provides the necessary function to generate visualizations in
    the form of plots or animations from water network data.

    Given a :class:`~epyt_flow.simulation.scenario_simulator.ScenarioSimulator`
    object, this class provides the necessary functions to plot the network
    topology and to color hydraulic elements according to simulation data. The
    resulting plot can then be displayed or saved.

    Attributes
    ----------
    __scenario : :class:`~epyt_flow.simulation.scenario_simulator.ScenarioSimulator`
        ScenarioSimulator object containing the network topology and
        configurations to obtain the simulation data which should be displayed.
    fig : `matplotlib.pyplot.Figure <https://matplotlib.org/stable/api/_as_gen/matplotlib.figure.Figure.html#matplotlib.figure.Figure>`_ or None
        Figure object used for plotting, created and customized by calling the
        methods of this class, initialized as None.
    ax : `~matplotlib.axes.Axes <https://matplotlib.org/stable/api/_as_gen/matplotlib.axes.Axes.html#matplotlib-axes-axes>`_ or None
        The axes for plotting, initialized as None.
    scada_data : :class:`~epyt_flow.simulation.scada.scada_data.ScadaData` or None
        SCADA data created by the ScenarioSimulator object, initialized as
        None.
    topology : :class:`~epyt_flow.topology.NetworkTopology`
        Topology object retrieved from the scenario, containing the structure
        of the water distribution network.
    color_scheme : :class:`~epyt_flow.visualization.visualization_utils.ColorScheme`
        Contains the selected ColorScheme for visualization.
    pipe_parameters : :class:`~epyt_flow.visualization.visualization_utils.EdgeObject`
        Class contains parameters for visualizing pipes in the correct format
        for drawing.
    junction_parameters : :class:`~epyt_flow.visualization.visualization_utils.JunctionObject`
        Class contains parameters for visualizing junctions in the correct
        format for drawing.
    tank_parameters : :class:`~epyt_flow.visualization.visualization_utils.JunctionObject`
        Class contains parameters for visualizing tanks in the correct format
        for drawing.
    reservoir_parameters : :class:`~epyt_flow.visualization.visualization_utils.JunctionObject`
        Class contains parameters for visualizing reservoirs in the correct
        format for
        drawing.
    valve_parameters : :class:`~epyt_flow.visualization.visualization_utils.JunctionObject`
        Class contains parameters for visualizing valves in the correct format
        for drawing.
    pump_parameters : :class:`~epyt_flow.visualization.visualization_utils.JunctionObject`
        Class contains parameters for visualizing pumps in the correct format
        for drawing.
    colorbars : `dict`
        A dictionary containing the necessary data for drawing the required
        colorbars.
    labels : `dict`
        A dictionary containing components as keys and drawing information for
        the labels as values.

    """

    def __init__(self, scenario: ScenarioSimulator,
                 color_scheme: ColorScheme = epyt_flow_colors) -> None:
        """
        Initializes the class with a given scenario, sets up the topology,
        SCADA data, and the classes containing parameters for visualizing
        various hydraulic components (pipes, junctions, tanks, reservoirs,
        valves, and pumps).

        Parameters
        ----------
        scenario : :class:`~epyt_flow.simulation.scenario_simulator.ScenarioSimulator`
            An instance of the `ScenarioSimulator` class, used to simulate and
            retrieve the system topology.
        color_scheme : :class:`~epyt_flow.visualization.visualization_utils.ColorScheme`
            Contains the selected ColorScheme for visualization. Default is
            EPYT_FLOW.

        Raises
        ------
        TypeError
            If `scenario` is not an instance of
            :class:`~epyt_flow.simulation.scenario_simulator.ScenarioSimulator`.
        """
        if not isinstance(scenario, ScenarioSimulator):
            raise TypeError("'scenario' must be an instance of " +
                            "'epyt_flow.simulation.ScenarioSimulator' " +
                            f"but not of '{type(scenario)}'")

        self.__scenario = scenario
        self.fig = None
        self.ax = None
        self.scada_data = None
        markers = Marker()
        self.topology = self.__scenario.get_topology()

        pos_dict = {x: self.topology.get_node_info(x)['coord'] for x in
                    self.topology.get_all_nodes()}

        self.color_scheme = color_scheme

        self.pipe_parameters = EdgeObject(
            [x[1] for x in self.topology.get_all_links()], pos_dict,
            self.color_scheme.pipe_color)
        self.junction_parameters = JunctionObject(
            self.topology.get_all_junctions(), pos_dict,
            node_color=self.color_scheme.node_color)
        self.tank_parameters = JunctionObject(self.topology.get_all_tanks(),
                                              pos_dict, node_size=100,
                                              node_shape=markers.tank,
                                              node_color=self.color_scheme.tank_color)
        self.reservoir_parameters = JunctionObject(
            self.topology.get_all_reservoirs(), pos_dict, node_size=100,
            node_shape=markers.reservoir,
            node_color=self.color_scheme.reservoir_color)
        self.valve_parameters = JunctionObject(self.topology.get_all_valves(),
                                               self._get_midpoints(
                                                   self.topology.get_all_valves()),
                                               node_size=50,
                                               node_shape=markers.valve,
                                               node_color=self.color_scheme.valve_color)
        self.pump_parameters = JunctionObject(self.topology.get_all_pumps(),
                                              self._get_midpoints(
                                                  self.topology.get_all_pumps()),
                                              node_size=50,
                                              node_shape=markers.pump,
                                              node_color=self.color_scheme.pump_color)

        self.colorbars = {}
        self.labels = {}
        self.masks = {}

    def _get_midpoints(self, elements: List[str]) -> dict[
        str, tuple[float, float]]:
        """
        Computes and returns the midpoints for drawing either valves or pumps
        in a water distribution network.

        For each element ID in the provided list, the method calculates the
        midpoint between its start and end nodes' coordinates.

        Parameters
        ----------
        elements : `list[str]`
            A list of element IDs (e.g., pump IDs, valve IDs) for which to
            compute the midpoints.

        Returns
        -------
        elements_dict : `dict`
            A dictionary where the keys are element IDs and the values are the
            corresponding midpoints, represented as 2D coordinates [x, y].
        """
        elements_pos_dict = {}
        for element in elements:
            if element in self.topology.pumps:
                start_node, end_node = self.topology.get_pump_info(element)[
                    'end_points']
            elif element in self.topology.valves:
                start_node, end_node = self.topology.get_valve_info(element)[
                    'end_points']
            else:
                raise ValueError(f"Unknown element '{element}'")
            start_pos = self.topology.get_node_info(start_node)['coord']
            end_pos = self.topology.get_node_info(end_node)['coord']
            pos = [(start_pos[0] + end_pos[0]) / 2,
                   (start_pos[1] + end_pos[1]) / 2]
            elements_pos_dict[element] = pos
        return elements_pos_dict

    def _get_next_frame(self, frame_number: int) -> None:
        """
        Draws the next frame of a water distribution network animation.

        This method updates a visualization animation with the hydraulic
        components colored according to the scada data corresponding to the
        current frame.

        Parameters
        ----------
        frame_number : `int`
            The current frame number used to retrieve the data corresponding to
            that frame
        """
        plt.clf()
        self.ax = self.fig.add_subplot(111)
        self.ax.axis('off')

        nxp.draw_networkx_edges(self.topology, ax=self.ax,
                                label='Pipes',
                                **self.pipe_parameters.get_frame(frame_number))
        nxp.draw_networkx_nodes(self.topology, ax=self.ax,
                                label='Junctions',
                                **self.junction_parameters.get_frame(
                                    frame_number))
        nxp.draw_networkx_nodes(self.topology, ax=self.ax,
                                label='Tanks',
                                **self.tank_parameters.get_frame(frame_number))
        nxp.draw_networkx_nodes(self.topology, ax=self.ax,
                                label='Reservoirs',
                                **self.reservoir_parameters.get_frame(
                                    frame_number))
        nxp.draw_networkx_nodes(
            self.topology, ax=self.ax,
            label='Valves', **self.valve_parameters.get_frame(frame_number))
        nxp.draw_networkx_nodes(
            self.topology,
            ax=self.ax, label='Pumps',
            **self.pump_parameters.get_frame(frame_number))

        for key, mask in self.masks.items():
            if key == 'nodes':
                nxp.draw_networkx_nodes(self.topology, ax=self.ax,
                                        **self.junction_parameters.get_frame_mask(
                                            mask,
                                            self.color_scheme.node_color))
            if key == 'pumps':
                nxp.draw_networkx_nodes(
                    self.topology,
                    ax=self.ax,
                    **self.pump_parameters.get_frame_mask(mask,
                                                          self.color_scheme.pump_color))
            if key == 'links':
                nxp.draw_networkx_edges(self.topology, ax=self.ax,
                                        **self.pipe_parameters.get_frame_mask(
                                            frame_number,
                                            self.color_scheme.pipe_color))
            if key == 'tanks':
                nxp.draw_networkx_nodes(self.topology, ax=self.ax,
                                        **self.tank_parameters.get_frame_mask(
                                            mask,
                                            self.color_scheme.tank_color))
            if key == 'valves':
                nxp.draw_networkx_nodes(
                    self.topology, ax=self.ax,
                    **self.valve_parameters.get_frame_mask(mask,
                                                           self.color_scheme.valve_color))

        self._draw_labels()
        self.ax.legend(fontsize=6)

        for colorbar_stats in self.colorbars.values():
            self.fig.colorbar(ax=self.ax, **colorbar_stats)

    def _interpolate_frames(self, num_inter_frames: int):
        """
        Interpolates intermediate values between frames using cubic spline
        interpolation for smoother animation.

        Parameters
        ----------
        num_inter_frames : `int`
            Number of total frames after interpolation.

        Returns
        -------
        num_inter_frames : `int`
            Number of total frames after interpolation.
        """
        for node_source in [self.junction_parameters, self.tank_parameters,
                            self.reservoir_parameters, self.valve_parameters,
                            self.pump_parameters]:
            node_source.interpolate(num_inter_frames)
        self.pipe_parameters.interpolate(num_inter_frames)

        return num_inter_frames

    def _draw_labels(self):
        """
        Method accesses the dict `self.labels` and draws all generated labels
        within.
        """
        for k, v in self.labels.items():
            if k in ['pipes']:
                nxp.draw_networkx_edge_labels(self.topology, ax=self.ax, **v)
                continue
            nxp.draw_networkx_labels(self.topology, ax=self.ax, **v)

    def _get_sensor_config_nodes_and_links(self):
        """
        Iterates through the sensor config and collects all nodes and links
        within, that have a sensor attached.

        Returns
        -------
        highlighted_links : `list`
            List of all links with sensors.
        highlighted_nodes : `list`
            List of all nodes with sensors.
        """
        highlighted_nodes = []
        highlighted_links = []

        sensor_config = self.__scenario.sensor_config
        highlighted_nodes += (sensor_config.pressure_sensors
                              + sensor_config.demand_sensors
                              + sensor_config.quality_node_sensors)
        highlighted_links += (sensor_config.flow_sensors
                              + sensor_config.quality_link_sensors)
        return highlighted_nodes, highlighted_links

    def add_labels(self, components: str or list or tuple = (),
                   font_size: int = 8):
        """
        Adds labels to hydraulic components according to the specified
        components.

        Parameters
        ----------
        components : `str` or `list` or `tuple`, default is ()
            Can either be 'all': all components, 'sensor_config': all nodes and
            pipes which have a sensor attached, or a list of the component
            names that are to be labeled: nodes, tanks, reservoirs, pipes,
            valves, pumps. If the list is empty, all nodes are labeled.
        font_size : `int`, default is 8
            Font size of the labels.
        """
        sc_nodes, sc_links = None, None
        if components == 'all':
            components = ['nodes', 'tanks', 'reservoirs', 'pipes', 'valves',
                          'pumps']
        elif components == 'sensor_config':
            components = ['nodes', 'tanks', 'reservoirs', 'pipes', 'valves',
                          'pumps']
            sc_nodes, sc_links = self._get_sensor_config_nodes_and_links()

        elif len(components) == 0:
            components = ['nodes']

        component_mapping = {
            'nodes': self.junction_parameters,
            'tanks': self.tank_parameters,
            'reservoirs': self.reservoir_parameters,
            'valves': self.valve_parameters,
            'pumps': self.pump_parameters
        }

        for component, parameters in component_mapping.items():
            if component in components:
                labels = {n: str(n) for n in parameters.nodelist if
                          sc_nodes is None or n in sc_nodes}
                self.labels[component] = {'pos': parameters.pos,
                                          'labels': labels,
                                          'font_size': font_size}
                if component in ['pumps', 'valves']:
                    self.labels[component]['verticalalignment'] = 'bottom'
        if 'pipes' in components:
            labels = {tuple(n[1]): n[0] for n in self.topology.get_all_links()
                      if sc_links is None or n[0] in sc_links}
            self.labels['pipes'] = {'pos': self.pipe_parameters.pos,
                                    'edge_labels': labels,
                                    'font_size': font_size}

    def show_animation(self, export_to_file: str = None,
                       return_animation: bool = False, duration: int = 5,
                       fps: int = 15, interpolate: bool = True) \
            -> Optional[FuncAnimation]:
        """
        Displays, exports, or returns an animation of a water distribution
        network over time.

        This method generates an animation of a network and either shows it or
        returns the :class:`~FuncAnimation` object. Optionally, the animation
        is saved to a file.

        Parameters
        ----------
        export_to_file : `str`, optional
            The file path where the animation should be saved, if provided.
            Default is `None`.
        return_animation : `bool`, optional
            If `True`, the animation object is returned. If `False`, the
            animation will be shown, but not returned. Default is `False`.
        duration : `int`, default is 5
            Duration of the animation in seconds.
        fps : `int`, default is 15
            Frames per seconds, is achieved through interpolation.
        interpolate : `bool`, default is True
            Whether to allow interpolating the sensor values or not. Necessary
            for fixed fps.

        Returns
        -------
        anim : :class:`~FuncAnimation` or None
            Returns the animation object if `return_animation` is `True`.
            Otherwise, returns `None`.
        """
        self.fig = plt.figure(figsize=(6.4, 4.8), dpi=200)

        total_frames = float('inf')
        for node_source in [self.junction_parameters, self.tank_parameters,
                            self.reservoir_parameters, self.valve_parameters,
                            self.pump_parameters]:
            if not isinstance(node_source.node_color, str) and len(
                    node_source.node_color) > 1:
                total_frames = min(total_frames, len(node_source.node_color))
        if hasattr(self.pipe_parameters, 'edge_color'):
            if not isinstance(self.pipe_parameters.edge_color, str) and len(
                    self.pipe_parameters.edge_color) > 1:
                total_frames = min(total_frames,
                                   len(self.pipe_parameters.edge_color))
        if hasattr(self.pipe_parameters, 'width'):
            if not isinstance(self.pipe_parameters.width, str) and len(
                    self.pipe_parameters.width) > 1:
                total_frames = min(total_frames,
                                   len(self.pipe_parameters.width))

        if total_frames == 0 or total_frames == float('inf'):
            raise RuntimeError("The color or resize functions must be called "
                               "with a time_step range (pit) > 1 to enable "
                               "animations")

        if interpolate:
            total_frames = self._interpolate_frames(fps * duration)

        anim = FuncAnimation(self.fig, self._get_next_frame,
                             frames=total_frames,
                             interval=round(duration * 100 / total_frames))

        if export_to_file is not None:
            anim.save(export_to_file, writer='ffmpeg', fps=fps)
        if return_animation:
            plt.close(self.fig)
            return anim
        plt.show()
        return None

    def show_plot(self, export_to_file: str = None,
                  suppress_plot: bool = False) -> None:
        """
        Displays a static plot of the water distribution network.

        This method generates a static plot of the water distribution network,
        visualizing pipes, junctions, tanks, reservoirs, valves, and pumps.
        The plot can be displayed and saved to a file.

        Parameters
        ----------
        export_to_file : `str`, optional
            The file path where the plot should be saved, if provided.
            Default is `None`.
        suppress_plot : `bool`, default is False
            If true, no plot is displayed after running this method.
        """
        self.fig = plt.figure(figsize=(6.4, 4.8), dpi=200)
        self._get_next_frame(0)

        if export_to_file is not None:
            plt.savefig(export_to_file, transparent=True, bbox_inches='tight',
                        dpi=900)
        if not suppress_plot:
            plt.show()
        else:
            plt.close(self.fig)
            plt.clf()

    def color_nodes(
            self, data: Optional[Union[ScadaData, np.ndarray]] = None,
            parameter: str = 'pressure', statistic: str = 'mean',
            pit: Optional[Union[int, Tuple[int, int]]] = None,
            species: str = None,
            colormap: str = 'viridis',
            intervals: Optional[Union[int, List[Union[int, float]]]] = None,
            conversion: Optional[dict] = None,
            show_colorbar: bool = False,
            use_sensor_data: bool = False) -> None:
        """
        Colors the nodes (junctions) in the water distribution network based on
        the SCADA data and the specified parameters.

        This method either takes or generates SCADA data, applies a statistic
        to the chosen parameter, optionally groups the results and prepares the
        results to be either displayed statically ot animated.

        Parameters
        ----------
        data : :class:`~epyt_flow.simulation.scada.scada_data.ScadaData` or
            `numpy.ndarray <https://numpy.org/doc/stable/reference/generated/numpy.ndarray.html>`_, optional
            The SCADA data object containing node data or a numpy array of the
            shape nodes*timesteps. If `None`, a simulation is run to generate
            SCADA data. Default is `None`.
        parameter : `str`, optional
            The node data to visualize. Must be 'pressure', 'demand', or
            'node_quality'. Default is 'pressure'.
        statistic : `str`, optional
            The statistic to calculate for the data. Can be 'mean', 'min',
            'max', or 'time_step'. Default is 'mean'.
        pit : `int`, `tuple(int, int)`, optional
            The point in time or range of time steps for the 'time_step'
            statistic. If a tuple is provided, it should contain two integers
            representing the start and end time steps. A tuple is necessary to
            process the data for the :meth:`~ScenarioVisualizer.show_animation`
            method. Default is `None`.
        species: `str`, optional
            Key of species. Only necessary for parameter
            'bulk_species_concentration'.
        colormap : `str`, optional
            The colormap to use for visualizing node values. Default is
            'viridis'.
        intervals : `int`, `list[int]` or `list[float]`, optional
            If provided, the data will be grouped into intervals. It can be an
            integer specifying the number of groups or a list of boundary
            points. Default is `None`.
        conversion : `dict`, optional
            A dictionary of conversion parameters to convert SCADA data units.
            Default is `None`.
        show_colorbar : `bool`, optional
            If `True`, a colorbar will be displayed on the plot to indicate the
            range of node values. Default is `False`.
        use_sensor_data : `bool`, optional
            If `True`, instead of using raw simulation data, the data recorded
            by the corresponding sensors in the system is used for the
            visualization. Note: Not all components may have a sensor attached
            and sensors may be subject to sensor faults or noise.

        Raises
        ------
        ValueError
            If the `parameter` is not one of 'pressure', 'demand', or
            'node_quality', or if `pit` is not correctly provided for the
            'time_step' statistic.

        """
        self.junction_parameters.cmap = colormap

        if data is not None:
            self.scada_data = data
        elif not self.scada_data:
            self.scada_data = self.__scenario.run_simulation()

        if conversion:
            self.scada_data = self.scada_data.convert_units(**conversion)

        # TODO: is there any way to make this look better (e.g. do a mapping somewhere??)
        if parameter == 'pressure':
            if use_sensor_data:
                values, self.masks[
                    'nodes'] = self.scada_data.get_data_pressures_as_node_features()
            else:
                values = self.scada_data.pressure_data_raw
        elif parameter == 'demand':
            if use_sensor_data:
                values, self.masks[
                    'nodes'] = self.scada_data.get_data_demands_as_node_features()
            else:
                values = self.scada_data.demand_data_raw
        elif parameter == 'node_quality':
            if use_sensor_data:
                values, self.masks[
                    'nodes'] = self.scada_data.get_data_nodes_quality_as_node_features()
            else:
                values = self.scada_data.node_quality_data_raw
        elif parameter == 'custom_data':
            # Custom should have the dimensions (timesteps, nodes)
            values = self.scada_data
        elif parameter == 'bulk_species_concentration':
            if not species:
                raise ValueError('Species must be set when using bulk_species_'
                                 'concentration.')
            if use_sensor_data:
                values, self.masks[
                    'nodes'] = self.scada_data.get_data_bulk_species_concentrations_as_node_features()
                self.masks['nodes'] = self.masks['nodes'][:,
                                      self.scada_data.sensor_config.bulk_species.index(
                                          species)]
                values = values[:, :,
                         self.scada_data.sensor_config.bulk_species.index(
                             species)]
            else:
                values = self.scada_data.bulk_species_node_concentration_raw[:,
                         self.scada_data.sensor_config.bulk_species.index(
                             species), :]
        else:
            raise ValueError(
                'Parameter must be pressure, demand, node_quality or custom_'
                'data.')

        if statistic == 'time_step' and isinstance(pit, tuple) and len(
                pit) == 2 and all(isinstance(i, int) for i in pit):
            rng = pit
            if pit[1] == -1:
                rng = (pit[0], values.shape[0])
            for frame in range(*rng):
                if frame > values.shape[0] - 1:
                    break
                self.junction_parameters.add_frame(statistic, values, frame,
                                                   intervals)
        else:
            self.junction_parameters.add_frame(statistic, values, pit,
                                               intervals)

        if show_colorbar:
            if statistic == 'time_step':
                label = str(parameter).capitalize() + ' at timestep ' + str(
                    pit)
            else:
                label = str(statistic).capitalize() + ' ' + str(
                    parameter).replace('_', ' ')
            self.colorbars['junctions'] = {'mappable': plt.cm.ScalarMappable(
                norm=mpl.colors.Normalize(
                    vmin=self.junction_parameters.vmin,
                    vmax=self.junction_parameters.vmax), cmap=colormap),
                'label': label}

    def color_links(
            self, data: Optional[Union[ScadaData, np.ndarray]] = None,
            parameter: str = 'flow_rate', statistic: str = 'mean',
            pit: Optional[Union[int, Tuple[int, int]]] = None,
            species: str = None,
            colormap: str = 'coolwarm',
            intervals: Optional[Union[int, List[Union[int, float]]]] = None,
            conversion: Optional[dict] = None,
            show_colorbar: bool = False,
            use_sensor_data: bool = False) -> None:
        """
        Colors the links (pipes) in the water distribution network based on the
        SCADA data and the specified parameters.

        This method either takes or generates SCADA data, applies a statistic
        to the chosen parameter, optionally groups the results and prepares the
        results to be either displayed statically ot animated.

        Parameters
        ----------
        data : :class:`~epyt_flow.simulation.scada.scada_data.ScadaData` or
            `numpy.ndarray <https://numpy.org/doc/stable/reference/generated/numpy.ndarray.html>`_, optional
            The SCADA data object containing link data or a numpy array of the
            shape links*timesteps. If `None`, a simulation is run to generate
            SCADA data. Default is `None`.
        parameter : `str`, optional
            The link data to visualize. Options are 'flow_rate', 'velocity', or
            'status'. Default is 'flow_rate'.
        statistic : `str`, optional
            The statistic to calculate for the data. Can be 'mean', 'min',
            'max', or 'time_step'. Default is 'mean'.
        pit : `int` or `tuple(int, int)`, optional
            The point in time or range of time steps for the 'time_step'
            statistic. If a tuple is provided, it should contain two integers
            representing the start and end time steps. A tuple is necessary to
            process the data for the :func:`~ScenarioVisualizer.show_animation`
            method. Default is `None`.
        species: `str`, optional
            Key of species. Only necessary for parameter
            'bulk_species_concentration'.
        colormap : `str`, optional
            The colormap to use for visualizing link values. Default is
            'coolwarm'.
        intervals : `int`, `list[int]`, `list[float]`, optional
            If provided, the data will be grouped into intervals. It can be an
            integer specifying the number of groups or a list of boundary
            points. Default is `None`.
        conversion : `dict`, optional
            A dictionary of conversion parameters to convert SCADA data units.
            Default is `None`.
        show_colorbar : `bool`, optional
            If `True`, a colorbar will be displayed on the plot to indicate the
            range of values. Default is `False`.
        use_sensor_data : `bool`, optional
            If `True`, instead of using raw simulation data, the data recorded
            by the corresponding sensors in the system is used for the
            visualization. Note: Not all components may have a sensor attached
            and sensors may be subject to sensor faults or noise.

        Raises
        ------
        ValueError
            If `parameter` is not a valid link data parameter or if `pit` is
            incorrectly provided for the 'time_step' statistic.

        """
        sim_length = None

        if data is not None:
            self.scada_data = data
            if not isinstance(self.scada_data, ScadaData):
                sim_length = self.scada_data.shape[0]
        elif not self.scada_data:
            self.scada_data = self.__scenario.run_simulation()

        if conversion:
            self.scada_data = self.scada_data.convert_units(**conversion)

        self.pipe_parameters.edge_cmap = mpl.colormaps[colormap]

        if sim_length is None:
            sim_length = self.scada_data.sensor_readings_time.shape[0]

        if statistic == 'time_step' and isinstance(pit, tuple) and len(
                pit) == 2 and all(isinstance(i, int) for i in pit):
            rng = pit
            if pit[1] == -1:
                rng = (pit[0], sim_length)
            for frame in range(*rng):
                if frame >= sim_length:
                    break
                self.pipe_parameters.add_frame(self.topology, 'edge_color',
                                               self.scada_data, parameter,
                                               statistic, frame, species,
                                               intervals, use_sensor_data)
        else:
            self.pipe_parameters.add_frame(self.topology, 'edge_color',
                                           self.scada_data, parameter,
                                           statistic, pit, species, intervals,
                                           use_sensor_data)

        if hasattr(self.pipe_parameters, 'mask'):
            self.masks['links'] = self.pipe_parameters.mask

        if show_colorbar:
            if statistic == 'time_step':
                label = (str(parameter).capitalize().replace('_', ' ')
                         + ' at timestep ' + str(pit))
            else:
                label = str(statistic).capitalize() + ' ' + str(
                    parameter).replace('_', ' ')
            self.colorbars['pipes'] = {'mappable': plt.cm.ScalarMappable(
                norm=mpl.colors.Normalize(
                    vmin=self.pipe_parameters.edge_vmin,
                    vmax=self.pipe_parameters.edge_vmax), cmap=colormap),
                'label': label}

    def color_pumps(
            self, data: Optional[Union[ScadaData, np.ndarray]] = None,
            parameter: str = 'efficiency', statistic: str = 'mean',
            pit: Optional[Union[int, Tuple[int]]] = None,
            intervals: Optional[Union[int, List[Union[int, float]]]] = None,
            colormap: str = 'viridis', show_colorbar: bool = False,
            use_sensor_data: bool = False) -> None:
        """
        Colors the pumps in the water distribution network based on SCADA data
        and the specified parameters.

        This method either takes or generates SCADA data, applies a statistic
        to the chosen parameter, optionally groups the results and prepares the
        results to be either displayed statically ot animated.

        Parameters
        ----------
        data : :class:`~epyt_flow.simulation.scada.scada_data.ScadaData` or
            `numpy.ndarray <https://numpy.org/doc/stable/reference/generated/numpy.ndarray.html>`_, optional
            The SCADA data object containing pump data or a numpy array of the
            shape pumps*timesteps. If `None`, a simulation is run to generate
            SCADA data. Default is `None`.
        parameter : `str`, optional
            The pump data to visualize. Must be 'efficiency',
            'energy_consumption', or 'state'. Default is 'efficiency'.
        statistic : `str`, optional
            The statistic to calculate for the data. Can be 'mean', 'min',
            'max', or 'time_step'. Default is 'mean'.
        pit : `int`, `tuple(int, int)`, optional
            The point in time or range of time steps for the 'time_step'
            statistic. If a tuple is provided, it should contain two integers
            representing the start and end time steps. A tuple is necessary to
            process the data for the :meth:`~ScenarioVisualizer.show_animation`
            method. Default is `None`.
        intervals : `int`, `list[int]`, `list[float]`, optional
            If provided, the data will be grouped into intervals. It can be an
            integer specifying the number of groups or a list of boundary
            points. Default is `None`.
        colormap : `str`, optional
            The colormap to use for visualizing pump values. Default is
            'viridis'.
        show_colorbar : `bool`, optional
            If `True`, a colorbar will be displayed on the plot to indicate the
            range of pump values. Default is `False`.
        use_sensor_data : `bool`, optional
            If `True`, instead of using raw simulation data, the data recorded
            by the corresponding sensors in the system is used for the
            visualization. Note: Not all components may have a sensor attached
            and sensors may be subject to sensor faults or noise.

        Raises
        ------
        ValueError
            If the `parameter` is not one of 'efficiency',
            'energy_consumption', or 'state', or if `pit` is not correctly
            provided for the 'time_step' statistic.

        """

        self.pump_parameters.cmap = colormap

        if data is not None:
            self.scada_data = data
        elif not self.scada_data:
            self.scada_data = self.__scenario.run_simulation()

        if parameter == 'efficiency':
            if use_sensor_data:
                values, self.masks[
                    'pumps'] = self.scada_data.get_data_pumps_efficiency_as_node_features()
            else:
                values = self.scada_data.pumps_efficiency_data_raw
        elif parameter == 'energy_consumption':
            if use_sensor_data:
                values, self.masks[
                    'pumps'] = self.scada_data.get_data_pumps_energyconsumption_as_node_features()
            else:
                values = self.scada_data.pumps_energyconsumption_data_raw
        elif parameter == 'state':
            if use_sensor_data:
                values, self.masks[
                    'pumps'] = self.scada_data.get_data_pumps_state_as_node_features()
            else:
                values = self.scada_data.pumps_state_data_raw
        elif parameter == 'custom_data':
            values = self.scada_data
        else:
            raise ValueError(
                'Parameter must be efficiency, energy_consumption, state or custom_data')

        if statistic == 'time_step' and isinstance(pit, tuple) and len(
                pit) == 2 and all(isinstance(i, int) for i in pit):
            rng = pit
            if pit[1] == -1:
                rng = (pit[0], values.shape[0])
            for frame in range(*rng):
                if frame > values.shape[0] - 1:
                    break
                self.pump_parameters.add_frame(statistic, values, frame,
                                               intervals)
        else:
            self.pump_parameters.add_frame(statistic, values, pit, intervals)

        if show_colorbar:
            if statistic == 'time_step':
                label = str(parameter).capitalize().replace(
                    '_', ' ') + ' at timestep ' + str(pit)
            else:
                label = str(statistic).capitalize() + ' ' + str(
                    parameter).replace('_', ' ')
            self.colorbars['pumps'] = {'mappable': plt.cm.ScalarMappable(
                norm=mpl.colors.Normalize(vmin=self.pump_parameters.vmin,
                                          vmax=self.pump_parameters.vmax),
                cmap=colormap), 'label': label}

    def color_tanks(
            self, data: Optional[Union[ScadaData, np.ndarray]] = None,
            statistic: str = 'mean',
            pit: Optional[Union[int, Tuple[int, int]]] = None,
            intervals: Optional[Union[int, List[Union[int, float]]]] = None,
            colormap: str = 'viridis', show_colorbar: bool = False,
            use_sensor_data: bool = False) -> None:
        """
        Colors the tanks in the water distribution network based on the SCADA
        tank volume data and the specified statistic.

        This method either takes or generates SCADA data, applies a statistic
        to the tank volume data, optionally groups the results and prepares
        them to be either displayed statically ot animated.

        Parameters
        ----------
        data : :class:`~epyt_flow.simulation.scada.scada_data.ScadaData` or
            `numpy.ndarray <https://numpy.org/doc/stable/reference/generated/numpy.ndarray.html>`_, optional
            The SCADA data object containing tank data or a numpy array of the
            shape tanks*timesteps. If `None`, a simulation is run to generate
            SCADA data. Default is `None`.
        statistic : `str`, optional
            The statistic to calculate for the data. Can be 'mean', 'min',
            'max', or 'time_step'. Default is 'mean'.
        pit : `int`, `tuple(int, int)`, optional
            The point in time or range of time steps for the 'time_step'
            statistic. If a tuple is provided, it should contain two integers
            representing the start and end time steps. A tuple is necessary to
            process the data for the :meth:`~ScenarioVisualizer.show_animation`
            method. Default is `None`.
        intervals : `int`, `list[int]`, `list[float]`, optional
            If provided, the data will be grouped into intervals. It can be an
            integer specifying the number of groups or a list of boundary
            points. Default is `None`.
        colormap : `str`, optional
            The colormap to use for visualizing tank values. Default is
            'viridis'.
        show_colorbar : `bool`, optional
            If `True`, a colorbar will be displayed on the plot to indicate the
            range of tank volume values. Default is `False`.
        use_sensor_data : `bool`, optional
            If `True`, instead of using raw simulation data, the data recorded
            by the corresponding sensors in the system is used for the
            visualization. Note: Not all components may have a sensor attached
            and sensors may be subject to sensor faults or noise.

        Raises
        ------
        ValueError
            If `pit` is not correctly provided for the 'time_step' statistic.

        """
        self.tank_parameters.cmap = colormap

        if data is not None:
            self.scada_data = data
        elif not self.scada_data:
            self.scada_data = self.__scenario.run_simulation()

        if isinstance(self.scada_data, ScadaData):
            if use_sensor_data:
                values, self.masks[
                    'tanks'] = self.scada_data.get_data_tanks_water_volume_as_node_features()
            else:
                values = self.scada_data.tanks_volume_data_raw
            parameter = 'tank volume'
        else:
            values = self.scada_data
            parameter = 'custom data'

        if statistic == 'time_step' and isinstance(pit, tuple) and len(
                pit) == 2 and all(isinstance(i, int) for i in pit):
            rng = pit
            if pit[1] == -1:
                rng = (pit[0], values.shape[0])
            for frame in range(*rng):
                if frame > values.shape[0] - 1:
                    break
                self.tank_parameters.add_frame(statistic, values, frame,
                                               intervals)
        else:
            self.tank_parameters.add_frame(statistic, values, pit, intervals)

        if show_colorbar:
            if statistic == 'time_step':
                label = parameter.capitalize() + ' at timestep ' + str(pit)
            else:
                label = str(statistic).capitalize() + ' ' + parameter
            self.colorbars['tanks'] = {'mappable': plt.cm.ScalarMappable(
                norm=mpl.colors.Normalize(vmin=self.tank_parameters.vmin,
                                          vmax=self.tank_parameters.vmin),
                cmap=colormap), 'label': label}

    def color_valves(
            self, data: Optional[Union[ScadaData, np.ndarray]] = None,
            statistic: str = 'mean',
            pit: Optional[Union[int, Tuple[int, int]]] = None,
            intervals: Optional[Union[int, List[Union[int, float]]]] = None,
            colormap: str = 'viridis', show_colorbar: bool = False,
            use_sensor_data: bool = False) -> None:
        """
        Colors the valves in the water distribution network based on SCADA
        valve state data and the specified statistic.

        This method either takes or generates SCADA data, applies a statistic
        to the valve state data, optionally groups the results and prepares
        them to be either displayed statically ot animated.

        Parameters
        ----------
        data : :class:`~epyt_flow.simulation.scada.scada_data.ScadaData` or
            `numpy.ndarray <https://numpy.org/doc/stable/reference/generated/numpy.ndarray.html>`_, optional
            The SCADA data object containing valve data or a numpy array of the
            shape valves*timesteps. If `None`, a simulation is run to generate
            SCADA data. Default is `None`.
        statistic : `str`, optional
            The statistic to calculate for the data. Can be 'mean', 'min',
            'max', or 'time_step'. Default is 'mean'.
        pit : `int`, `tuple(int)`, optional
            The point in time or range of time steps for the 'time_step'
            statistic. If a tuple is provided, it should contain two integers
            representing the start and end time steps. A tuple is necessary to
            process the data for the :meth:`~ScenarioVisualizer.show_animation`
            method. Default is `None`.
        intervals : `int`, `list[int]`, `list[float]`, optional
            If provided, the data will be grouped into intervals. It can be an
            integer specifying the number of groups or a list of
            boundary points. Default is `None`.
        colormap : `str`, optional
            The colormap to use for visualizing valve state values. Default is
            'viridis'.
        show_colorbar : `bool`, optional
            If `True`, a colorbar will be displayed on the plot to indicate the
            range of valve state values. Default is `False`.
        use_sensor_data : `bool`, optional
            If `True`, instead of using raw simulation data, the data recorded
            by the corresponding sensors in the system is used for the
            visualization. Note: Not all components may have a sensor attached
            and sensors may be subject to sensor faults or noise.

        Raises
        ------
        ValueError
            If `pit` is not correctly provided for the 'time_step' statistic.

        """

        self.valve_parameters.cmap = colormap

        if data is not None:
            self.scada_data = data
        elif not self.scada_data:
            self.scada_data = self.__scenario.run_simulation()

        if isinstance(self.scada_data, ScadaData):
            if use_sensor_data:
                values, self.masks[
                    'valves'] = self.scada_data.get_data_valves_state_as_node_features()
            else:
                values = self.scada_data.valves_state_data_raw
            parameter = 'valve state'
        else:
            values = self.scada_data
            parameter = 'custom data'

        if statistic == 'time_step' and isinstance(pit, tuple) and len(
                pit) == 2 and all(isinstance(i, int) for i in pit):
            rng = pit
            if pit[1] == -1:
                rng = (pit[0], values.shape[0])
            for frame in range(*rng):
                if frame > values.shape[0] - 1:
                    break
                self.valve_parameters.add_frame(statistic, values, frame,
                                                intervals)
        else:
            self.valve_parameters.add_frame(statistic, values, pit,
                                            intervals)

        if show_colorbar:
            if statistic == 'time_step':
                label = parameter.capitalize() + ' at timestep ' + str(pit)
            else:
                label = str(statistic).capitalize() + ' ' + 'valve state'
            self.colorbars['valves'] = {'mappable': plt.cm.ScalarMappable(
                norm=mpl.colors.Normalize(vmin=self.valve_parameters.vmin,
                                          vmax=self.valve_parameters.vmax),
                cmap=colormap), 'label': label}

    def resize_links(
            self, data: Optional[Union[ScadaData, np.ndarray]] = None,
            parameter: str = 'flow_rate', statistic: str = 'mean',
            line_widths: Tuple[int, int] = (1, 2),
            pit: Optional[Union[int, Tuple[int, int]]] = None,
            species: str = None,
            intervals: Optional[Union[int, List[Union[int, float]]]] = None,
            conversion: Optional[dict] = None,
            use_sensor_data: bool = False) -> None:
        """
        Resizes the width of the links (pipes) in the water distribution
        network based on SCADA data and the specified parameters.

        This method either takes or generates SCADA data, applies a statistic,
        optionally groups the results and prepares them to be either displayed
        statically ot animated as link width.

        Parameters
        ----------
        data : :class:`~epyt_flow.simulation.scada.scada_data.ScadaData` or
            `numpy.ndarray <https://numpy.org/doc/stable/reference/generated/numpy.ndarray.html>`_, optional
            The SCADA data object containing link data or a numpy array of the
            shape links*timesteps. If `None`, a simulation is run to generate
            SCADA data. Default is `None`.
        parameter : `str`, optional
            The data used to resize to. Default is 'flow_rate'.
        statistic : `str`, optional
            The statistic to calculate for the data. Can be 'mean', 'min',
            'max', or 'time_step'. Default is 'mean'.
        line_widths : `tuple(int, int)`, optional
            A tuple specifying the range of line widths to use when resizing
            links based on the data. Default is (1, 2).
        pit : `int` or `tuple(int, int)`, optional
            The point in time or range of time steps for the 'time_step'
            statistic. If a tuple is provided, it should contain two integers
            representing the start and end time steps. A tuple is necessary to
            process the data for the :meth:`~ScenarioVisualizer.show_animation`
            method. Default is `None`.
        species: `str`, optional
            Key of species. Only necessary for parameter
            'bulk_species_concentration'.
        intervals : `int` or `list[int]` or `list[float]`, optional
            If provided, the data will be grouped into intervals. It can be an
            integer specifying the number of groups or a list of boundary
            points. Default is `None`.
        conversion : `dict`, optional
            A dictionary of conversion parameters to convert SCADA data units.
            Default is `None`.
        use_sensor_data : `bool`, optional
            If `True`, instead of using raw simulation data, the data recorded
            by the corresponding sensors in the system is used for the
            visualization. Note: Not all components may have a sensor attached
            and sensors may be subject to sensor faults or noise.
        """
        sim_length = None

        if data is not None:
            self.scada_data = data
            if not isinstance(self.scada_data, ScadaData):
                sim_length = self.scada_data.shape[0]
        elif not self.scada_data:
            self.scada_data = self.__scenario.run_simulation()

        if conversion:
            self.scada_data = self.scada_data.convert_units(**conversion)

        if sim_length is None:
            sim_length = self.scada_data.sensor_readings_time.shape[0]

        if statistic == 'time_step' and isinstance(pit, tuple) and len(
                pit) == 2 and all(isinstance(i, int) for i in pit):
            rng = pit
            if pit[1] == -1:
                rng = (pit[0], sim_length)
            for frame in range(*rng):
                if frame >= sim_length:
                    break
                self.pipe_parameters.add_frame(self.topology, 'edge_width',
                                               self.scada_data, parameter,
                                               statistic, frame, species,
                                               intervals, use_sensor_data)
        else:
            self.pipe_parameters.add_frame(self.topology, 'edge_width',
                                           self.scada_data, parameter,
                                           statistic, pit, species, intervals,
                                           use_sensor_data)
        self.pipe_parameters.rescale_widths(line_widths)

    def hide_nodes(self) -> None:
        """
        Hides all nodes (junctions) in the water distribution network
        visualization.

        This method clears the node list from the `junction_parameters`
        class, effectively removing all nodes from view in the current
        visualization.
        """
        self.junction_parameters.nodelist = []

    def highlight_sensor_config(self) -> None:
        """
        Highlights nodes and links that have sensors in the sensor_config in
        the water distribution network visualization.

        This method identifies nodes and links equipped with different types of
        sensors from the
        :class:`~epyt_flow.simulation.sensor_config.SensorConfig` and
        updates their visual appearance. Nodes with sensors are highlighted
        with an orange border, while links with sensors are displayed with a
        dashed line style.
        """
        highlighted_nodes, highlighted_links = self._get_sensor_config_nodes_and_links()

        node_edges = [
            (17, 163, 252) if node in highlighted_nodes else (0, 0, 0) for node
            in self.topology]
        pipe_style = ['dashed' if link[0] in highlighted_links else 'solid' for
                      link in self.topology.get_all_links()]

        self.junction_parameters.add_attributes(
            {'linewidths': 1, 'edgecolors': node_edges})
        self.pipe_parameters.add_attributes({'style': pipe_style})
